"""

Copyright 2019 European Union

Licensed under the EUPL, Version 1.2 or as soon they will be approved by the European Commission  subsequent versions of the EUPL (the "Licence");

You may not use this work except in compliance with the Licence.
You may obtain a copy of the Licence at:

https://joinup.ec.europa.eu/sites/default/files/inline-files/EUPL%20v1_2%20EN(1).txt

Unless required by applicable law or agreed to in writing, software distributed under the Licence is distributed on an "AS IS" basis,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the Licence for the specific language governing permissions and limitations under the Licence.

"""
from __future__ import print_function, absolute_import

from pcraster import lddmask, accuflux, boolean, downstream, pit, path, lddrepair, ifthenelse, cover, nominal, uniqueid, \
    catchment, upstream

import numpy as np

from .lakes import lakes
from .reservoir import reservoir
from .polder import polder
from .inflow import inflow
from .transmission import transmission
from .kinematic_wave_parallel import kinematicWave, kwpt

from ..global_modules.settings import LisSettings, MaskInfo
from ..global_modules.add1 import loadmap, loadmap_base, compressArray, decompress
from . import HydroModule


class routing(HydroModule):

    """
    # ************************************************************
    # ***** ROUTING      *****************************************
    # ************************************************************
    """
    input_files_keys = {'all': ['beta', 'ChanLength', 'Ldd', 'Channels', 'ChanGrad', 'ChanGradMin',
                                'CalChanMan', 'ChanMan', 'ChanBottomWidth', 'ChanDepthThreshold',
                                'ChanSdXdY', 'TotalCrossSectionAreaInitValue', 'PrevDischarge'],
                        'SplitRouting': ['CrossSection2AreaInitValue', 'PrevSideflowInitValue', 'CalChanMan2'],
                        'dynamicWave': ['ChannelsDynamic']}
    module_name = 'Routing'

    def __init__(self, routing_variable):
        self.var = routing_variable

        self.lakes_module = lakes(self.var)
        self.reservoir_module = reservoir(self.var)
        self.polder_module = polder(self.var)
        self.inflow_module = inflow(self.var)
        self.transmission_module = transmission(self.var)

# --------------------------------------------------------------------------
# --------------------------------------------------------------------------
    def initial(self):
        """ initial part of the routing module
        """
        maskinfo = MaskInfo.instance()
        self.var.avgdis = maskinfo.in_zero()
        self.var.Beta = loadmap('beta')
        self.var.InvBeta = 1 / self.var.Beta
        # Inverse of beta for kinematic wave
        self.var.ChanLength = loadmap('ChanLength').astype(float)
        self.var.InvChanLength = 1 / self.var.ChanLength
        # Inverse of channel length [1/m]

        self.var.NoRoutSteps = int(np.maximum(1, round(self.var.DtSec / self.var.DtSecChannel,0)))
        # Number of sub-steps based on value of DtSecChannel,
        # or 1 if DtSec is smaller than DtSecChannel
        settings = LisSettings.instance()
        option = settings.options
        if option['InitLisflood']:
            self.var.NoRoutSteps = 1
            # InitLisflood is used!
            # so channel routing step is the same as the general time step
        self.var.DtRouting = self.var.DtSec / self.var.NoRoutSteps
        # Corresponding sub-timestep (seconds)
        self.var.InvDtRouting = 1 / self.var.DtRouting
        self.var.InvNoRoutSteps = 1 / float(self.var.NoRoutSteps)
        # inverse for faster calculation inside the dynamic section

# -------------------------- LDD

        self.var.Ldd = lddmask(loadmap('Ldd', pcr=True, lddflag=True), self.var.MaskMap)
        # Cut ldd to size of MaskMap (NEW, 29/9/2004)
        # Prevents 'unsound' ldd if MaskMap covers sub-area of ldd

        # Count (inverse of) upstream area for each pixel
        # Needed if we want to calculate average values of variables
        # upstream of gauge locations

        self.var.UpArea = accuflux(self.var.Ldd, self.var.PixelAreaPcr)
        # Upstream contributing area for each pixel
        # Note that you might expext that values of UpArea would be identical to
        # those of variable CatchArea (see below) at the outflow points.
        # This is NOT actually the case, because outflow points are shifted 1
        # cell in upstream direction in the calculation of CatchArea!
        self.var.InvUpArea = 1 / self.var.UpArea
        # Calculate inverse, so we can multiply in dynamic (faster than divide)

        self.var.IsChannelPcr = boolean(loadmap('Channels', pcr=True))
        self.var.IsChannel = np.bool8(compressArray(self.var.IsChannelPcr))
        # Identify channel pixels
        self.var.IsChannelKinematic = self.var.IsChannel.copy()
        # Identify kinematic wave channel pixels
        # (identical to IsChannel, unless dynamic wave is used, see below)
        self.var.IsStructureKinematic = np.bool8(maskinfo.in_zero())

        # Map that identifies special inflow/outflow structures (reservoirs, lakes) within the
        # kinematic wave channel routing. Set to (dummy) value of zero modified in reservoir and lake
        # routines (if those are used)
        LddChan = lddmask(self.var.Ldd, self.var.IsChannelPcr)
        # ldd for Channel network
        self.var.MaskMap = boolean(self.var.Ldd)
        # Use boolean version of Ldd as calculation mask
        # (important for correct mass balance check
        # any water generated outside of Ldd won't reach
        # channel anyway)
        self.var.LddToChan = lddrepair(ifthenelse(self.var.IsChannelPcr, 5, self.var.Ldd))
        # Routing of runoff (incl. ground water)en
        AtOutflow = boolean(pit(self.var.Ldd))
        # find outlet points...

        if option['dynamicWave']:
            IsChannelDynamic = boolean(loadmap('ChannelsDynamic', pcr=True))
            # Identify channel pixels where dynamic wave is used
            self.var.IsChannelKinematic = (self.var.IsChannelPcr == 1) & (IsChannelDynamic == 0)
            # Identify (update) channel pixels where kinematic wave is used
            self.var.LddKinematic = lddmask(self.var.Ldd, self.var.IsChannelKinematic)
            # Ldd for kinematic wave: ends (pit) just before dynamic stretch

            # Following statements produce an ldd network that connects the pits in
            # LddKinematic to the nearest downstream dynamic wave pixel

            self.var.AtLastPoint = (downstream(self.var.Ldd, AtOutflow) == 1) & (AtOutflow != 1) & self.var.IsChannelPcr

            # NEW 23-6-2005
            # Dynamic wave routine gives no outflow out of pits, so we calculate this
            # one cell upstream (WvD)
            # (implies that most downstream cell is not taken into account in mass balance
            # calculations, even if dyn wave is not used)
            # Only include points that are on a channel (otherwise some small 'micro-catchments'
            # are included, for which the mass balance cannot be calculated
            # properly)

        else:
            self.var.LddKinematic = LddChan
            # No dynamic wave, so kinematic ldd equals channel ldd
            self.var.AtLastPoint = AtOutflow
            self.var.AtLastPointC = np.bool8(compressArray(self.var.AtLastPoint))
            # assign unique identifier to each of them
        maskinfo = MaskInfo.instance()
        lddC = compressArray(self.var.LddKinematic)
        inAr = decompress(np.arange(maskinfo.info.mapC[0], dtype="int32"))
        # giving a number to each non missing pixel as id
        self.var.downstruct = (compressArray(downstream(self.var.LddKinematic, inAr))).astype("int32")
        # each upstream pixel gets the id of the downstream pixel
        self.var.downstruct[lddC == 5] = maskinfo.info.mapC[0]
        # all pits gets a high number
        # upstream function in numpy

        OutflowPoints = nominal(uniqueid(self.var.AtLastPoint))
        # and assign unique identifier to each of them
        self.var.Catchments = (compressArray(catchment(self.var.Ldd, OutflowPoints))).astype(np.int32)
        CatchArea = np.bincount(self.var.Catchments, weights=self.var.PixelArea)[self.var.Catchments]
        # define catchment for each outflow point
        # Compute area of each catchment [m2]
        # Note: in earlier versions this was calculated using the "areaarea" function,
        # changed to "areatotal" in order to enable handling of grids with spatially
        # variable cell areas (e.g. lat/lon grids)
        self.var.InvCatchArea = 1 / CatchArea
        # inverse of catchment area [1/m2]

        # ************************************************************
        # ***** CHANNEL GEOMETRY  ************************************
        # ************************************************************

        self.var.ChanGrad = np.maximum(loadmap('ChanGrad'), loadmap('ChanGradMin'))
        # avoid calculation of Alpha using ChanGrad=0: this creates MV!
        self.var.CalChanMan = loadmap('CalChanMan')
        self.var.ChanMan = self.var.CalChanMan * loadmap('ChanMan')
        # Manning's n is multiplied by ChanManCal
        # enables calibration for peak timing
        self.var.ChanBottomWidth = loadmap('ChanBottomWidth')
        ChanDepthThreshold = loadmap('ChanDepthThreshold')
        ChanSdXdY = loadmap('ChanSdXdY')
        self.var.ChanUpperWidth = self.var.ChanBottomWidth + 2 * ChanSdXdY * ChanDepthThreshold
        # Channel upper width [m]
        self.var.TotalCrossSectionAreaBankFull = 0.5 * \
            ChanDepthThreshold * (self.var.ChanUpperWidth + self.var.ChanBottomWidth)
        # Area (sq m) of bank full discharge cross section [m2]
        # (trapezoid area equation)
        TotalCrossSectionAreaHalfBankFull = 0.5 * self.var.TotalCrossSectionAreaBankFull
        # Cross-sectional area at half bankfull [m2]
        # This can be used to initialise channel flow (see below)

        TotalCrossSectionAreaInitValue = loadmap('TotalCrossSectionAreaInitValue')
        self.var.TotalCrossSectionArea = np.where(TotalCrossSectionAreaInitValue == -9999, TotalCrossSectionAreaHalfBankFull, TotalCrossSectionAreaInitValue)
        # Total cross-sectional area [m2]: if initial value in binding equals -9999 the value at half bankfull is used,
        # otherwise TotalCrossSectionAreaInitValue (typically end map from previous simulation)

        if option['SplitRouting']:
            # in_zero = maskinfo.in_zero()
            CrossSection2AreaInitValue = loadmap('CrossSection2AreaInitValue')
            self.var.CrossSection2Area = np.where(CrossSection2AreaInitValue == -9999, maskinfo.in_zero(), CrossSection2AreaInitValue)
            # cross-sectional area [m2] for 2nd line of routing: if initial value in binding equals -9999 the value is set to 0
            # otherwise CrossSection2AreaInitValue (typically end map from previous simulation)

            PrevSideflowInitValue = loadmap('PrevSideflowInitValue')
            self.var.Sideflow1Chan = np.where(PrevSideflowInitValue == -9999, maskinfo.in_zero(), PrevSideflowInitValue)
            # sideflow from previous run for 1st line of routing: if initial value in binding equals -9999 the value is set to 0
            # otherwise PrevSideflowInitValue (typically end map from previous simulation)

        # ************************************************************
        # ***** CHANNEL ALPHA (KIN. WAVE)*****************************
        # ************************************************************
        # Following calculations are needed to calculate Alpha parameter in kinematic
        # wave. Alpha currently fixed at half of bankful depth (this may change in
        # future versions!)

        ChanWaterDepthAlpha = np.where(self.var.IsChannel, 0.5 * ChanDepthThreshold, 0.0)
        # Reference water depth for calculation of Alpha: half of bankfull
        self.var.ChanWettedPerimeterAlpha = self.var.ChanBottomWidth + 2 * \
            np.sqrt(np.square(ChanWaterDepthAlpha) + np.square(ChanWaterDepthAlpha * ChanSdXdY))
        # Channel wetted perimeter [m](Pythagoras)
        AlpTermChan = (self.var.ChanMan / (np.sqrt(self.var.ChanGrad))) ** self.var.Beta
        self.var.AlpPow = 2.0 / 3.0 * self.var.Beta

        self.var.ChannelAlpha = (AlpTermChan * (self.var.ChanWettedPerimeterAlpha ** self.var.AlpPow)).astype(float)
        self.var.InvChannelAlpha = 1 / self.var.ChannelAlpha
        # ChannelAlpha for kinematic wave

        # ************************************************************
        # ***** CHANNEL INITIAL DISCHARGE ****************************
        # ************************************************************

        self.var.ChanM3 = self.var.TotalCrossSectionArea * self.var.ChanLength
        # channel water volume [m3]
        self.var.ChanIniM3 = self.var.ChanM3.copy()
        self.var.ChanM3Kin = self.var.ChanIniM3.copy().astype(float)
        # Initialise water volume in kinematic wave channels [m3]
        self.var.ChanQKin = np.where(self.var.ChannelAlpha > 0, (self.var.TotalCrossSectionArea / self.var.ChannelAlpha) ** self.var.InvBeta, 0).astype(float)
        # Initialise discharge at kinematic wave pixels (note that InvBeta is
        # simply 1/beta, computational efficiency!)

        self.var.CumQ = maskinfo.in_zero()
        # ininialise sum of discharge to calculate average

# ************************************************************
# ***** CHANNEL INITIAL DYNAMIC WAVE *************************
# ************************************************************
        if option['dynamicWave']:
            pass
            # TODO !!!!!!!!!!!!!!!!!!!!

       #     lookchan = lookupstate(TabCrossSections, ChanCrossSections, ChanBottomLevel, self.var.ChanLength,
       #                            DynWaveConstantHeadBoundary + ChanBottomLevel)
       #     ChanIniM3 = ifthenelse(AtOutflow, lookchan, ChanIniM3)
            # Correct ChanIniM3 for constant head boundary in pit (only if
            # dynamic wave is used)
       #     ChanM3Dyn = ChanIniM3
            # Set volume of water in dynamic wave channel to initial value
            # (note that initial condition is expressed as a state in [m3] for the dynamic wave,
            # and as a rate [m3/s] for the kinematic wave (a bit confusing)

            # Estimate number of iterations needed in first time step (based on Courant criterium)
            # TO DO !!!!!!!!!!!!!!!!!!!!
        #    Potential = lookuppotential(
        #        TabCrossSections, ChanCrossSections, ChanBottomLevel, self.var.ChanLength, ChanM3Dyn)
            # Potential
        #    WaterLevelDyn = Potential - ChanBottomLevel
            # Water level [m above bottom level)
        #    WaveCelerityDyn = pcraster.sqrt(9.81 * WaterLevelDyn)
            # Dynamic wave celerity [m/s]
        #    CourantDynamic = self.var.DtSec * \
        #        (WaveCelerityDyn + 2) / self.var.ChanLength
            # Courant number for dynamic wave
            # We don't know the water velocity at this time so
            # we just guess it's 2 m/s (Odra tests show that flow velocity
            # is typically much lower than wave celerity, and 2 m/s is quite
            # high already so this gives a pretty conservative/safe estimate
            # for DynWaveIterations)
        #    DynWaveIterationsTemp = max(
        #        1, roundup(CourantDynamic / CourantDynamicCrit))
        #    DynWaveIterations = ordinal(mapmaximum(DynWaveIterationsTemp))
            # Number of sub-steps needed for required numerical
            # accuracy. Always greater than or equal to 1
            # (otherwise division by zero!)

            # TEST
            # If polder option is used, we need an estimate of the initial channel discharge, but we don't know this
            # for the dynamic wave pixels (since only initial state is known)! Try if this works (dyn wave flux based on zero inflow 1 iteration)
            # Note that resulting ChanQ is ONLY used in the polder routine!!!
            # Since we need instantaneous estimate at start of time step, a
            # ChanQM3Dyn is calculated for one single one-second time step!!!

        #    ChanQDyn = dynwaveflux(TabCrossSections,
        #                           ChanCrossSections,
        #                           LddDynamic,
        #                           ChanIniM3,
        #                           0.0,
        #                           ChanBottomLevel,
        #                           self.var.ChanMan,
        #                           self.var.ChanLength,
        #                           1,
        #                           1,
        #                           DynWaveBoundaryCondition)
            # Compute volume and discharge in channel after dynamic wave
            # ChanM3Dyn in [cu m]
            # ChanQDyn in [cu m / s]
        #    self.var.ChanQ = ifthenelse(
        #        IsChannelDynamic, ChanQDyn, self.var.ChanQKin)
            # Channel discharge: combine results of kinematic and dynamic wave
        else:

            # ***** NO DYNAMIC WAVE *************************
            # Dummy code if dynamic wave is not used, in which case ChanQ equals ChanQKin
            # (needed only for polder routine)

            PrevDischarge = loadmap('PrevDischarge')
            self.var.ChanQ = np.where(PrevDischarge == -9999, self.var.ChanQKin, PrevDischarge)
            # initialise channel discharge: cold start: equal to ChanQKin
            # [m3/s]

        # Initialising cumulative output variables
        # These are all needed to compute the cumulative mass balance error

        self.var.DischargeM3Out = maskinfo.in_zero()
        # cumulative discharge at outlet [m3]
        self.var.TotalQInM3 = maskinfo.in_zero()
        # cumulative inflow from inflow hydrographs [m3]
        self.var.sumDis = maskinfo.in_zero()
        self.var.sumInWB = maskinfo.in_zero()

    def initialSecond(self):
        """ initial part of the second channel routing module
        """
        settings = LisSettings.instance()
        option = settings.options
        binding = settings.binding

        self.var.ChannelAlpha2 = None  # default value, if split-routing is not active and only water is routed only in the main channel
        # ************************************************************
        # ***** CHANNEL INITIAL SPLIT UP IN SECOND CHANNEL************
        # ************************************************************
        if option['SplitRouting']:

            ChanMan2 = (self.var.ChanMan / self.var.CalChanMan) * loadmap('CalChanMan2')
            AlpTermChan2 = (ChanMan2 / (np.sqrt(self.var.ChanGrad))) ** self.var.Beta
            self.var.ChannelAlpha2 = (AlpTermChan2 * (self.var.ChanWettedPerimeterAlpha ** self.var.AlpPow)).astype(float)
            self.var.InvChannelAlpha2 = 1 / self.var.ChannelAlpha2
            # calculating second Alpha for second (virtual) channel

            if not(option['InitLisflood']):

                # use loadmap_base function as we don't want to cache avgdis it in the calibration
                self.var.QLimit = loadmap_base('AvgDis') * loadmap('QSplitMult')

                # Over bankful discharge starts at QLimit
                # lower discharge limit for second line of routing
                # set to mutiple of average discharge (map from prerun)
                # QSplitMult =2 is around 90 to 95% of Q

                self.var.M3Limit = self.var.ChannelAlpha * self.var.ChanLength * (self.var.QLimit ** self.var.Beta)
                # Water volume in bankful when over bankful discharge starts

                # QLimit should NOT be dependent on the NoRoutSteps (number of routing steps)
                # self.var.QLimit = self.var.QLimit / self.var.NoRoutSteps #original

                ###############################################
                # CM mod
                # TEMPORARY WORKAROUND FOR EFAS XDOM!!!!!!!!!!
                # This must be removed
                # self.var.QLimit = self.var.QLimit / 24.0
                ###############################################

                self.var.Chan2M3Start = self.var.ChannelAlpha2 * self.var.ChanLength * (self.var.QLimit ** self.var.Beta)
                # virtual amount of water in the channel through second line

                self.var.Chan2QStart = self.var.QLimit - compressArray(upstream(self.var.LddKinematic, decompress(self.var.QLimit)))
                # because kinematic routing with a low amount of discharge leads to long travel time:
                # Starting Q for second line is set to a higher value

                self.var.Chan2M3Kin = self.var.CrossSection2Area * self.var.ChanLength + self.var.Chan2M3Start
                self.var.ChanM3Kin = self.var.ChanM3 - self.var.Chan2M3Kin + self.var.Chan2M3Start
                
                self.var.ChanM3Kin = np.where((self.var.ChanM3Kin < 0.0) & (self.var.ChanM3Kin > -0.0000001),0.0,self.var.ChanM3Kin) # this line prevents the warmstart from failing in case of small numerical imprecisions when writing and reading the maps

                self.var.Chan2QKin = (self.var.Chan2M3Kin * self.var.InvChanLength * self.var.InvChannelAlpha2) ** (self.var.InvBeta)
                self.var.ChanQKin = (self.var.ChanM3Kin * self.var.InvChanLength * self.var.InvChannelAlpha) ** (self.var.InvBeta)

        # Initialise parallel kinematic wave router: main channel-only routing if self.var.ChannelAlpha2 is None; else split-routing(main channel + floodplains)
        maskinfo = MaskInfo.instance()
        self.river_router = kinematicWave(compressArray(self.var.LddKinematic), ~maskinfo.info.mask, self.var.ChannelAlpha,
                                          self.var.Beta, self.var.ChanLength, self.var.DtRouting,
                                          int(binding["numCPUs_parallelKinematicWave"]), alpha_floodplains=self.var.ChannelAlpha2)
        
        if option['InitLisflood'] and option['repMBTs']:          
            self.var.StorageStepINIT= self.var.ChanM3Kin 
            self.var.DischargeM3StructuresIni = maskinfo.in_zero()     
            if option['simulateReservoirs']:             
               self.var.StorageStepINIT += self.var.ReservoirStorageIniM3  
            if option['simulateLakes']:  
               self.var.StorageStepINIT += self.var.LakeStorageIniM3 
            self.var.StorageStepINIT = np.take(np.bincount(self.var.Catchments, weights=self.var.StorageStepINIT), self.var.Catchments) 
                
        if not option['InitLisflood'] and option['repMBTs']:  
           DisStructure = np.where(self.var.IsUpsOfStructureKinematicC, self.var.ChanQ * self.var.DtRouting, 0) 
           if not(option['SplitRouting']): 
            self.var.StorageStepINIT = self.var.ChanM3Kin      
            if option['simulateReservoirs']:                        
               self.var.StorageStepINIT += self.var.ReservoirStorageIniM3  
               DisStructure = np.where(self.var.IsUpsOfStructureKinematicC, self.var.ChanQ * self.var.DtRouting, 0)  
            if option['simulateLakes']:  
               self.var.StorageStepINIT += self.var.LakeStorageIniM3 
               DisStructure += np.where(compressArray(self.var.IsUpsOfStructureLake), 0.5 * self.var.ChanQ * self.var.DtRouting, 0) 
            self.var.DischargeM3StructuresIni = np.take(np.bincount(self.var.Catchments, weights=DisStructure), self.var.Catchments)      
           else:                             
            self.var.StorageStepINIT= self.var.ChanM3Kin+self.var.Chan2M3Kin-self.var.Chan2M3Start                       
            if option['simulateReservoirs']: 
               self.var.StorageStepINIT += self.var.ReservoirStorageIniM3
            if option['simulateLakes']:   
               self.var.StorageStepINIT += self.var.LakeStorageIniM3
            self.var.StorageStepINIT = np.take(np.bincount(self.var.Catchments, weights=self.var.StorageStepINIT), self.var.Catchments)    
# --------------------------------------------------------------------------
# --------------------------------------------------------------------------

    def dynamic(self, NoRoutingExecuted):
        """ dynamic part of the routing subtime module
        """
        settings = LisSettings.instance()
        option = settings.options
        binding = settings.binding

        if not(option['InitLisflood']):    # only with no InitLisflood
            self.lakes_module.dynamic_inloop(NoRoutingExecuted)
            self.reservoir_module.dynamic_inloop(NoRoutingExecuted)
            self.polder_module.dynamic_inloop()

        # End only with no Lisflood (no reservoirs, lakes and polder with
        # initLisflood)

        self.inflow_module.dynamic_inloop(NoRoutingExecuted)
        self.transmission_module.dynamic_inloop()

        # ************************************************************
        # ***** CHANNEL FLOW ROUTING: KINEMATIC WAVE  ****************
        # ************************************************************

        if not(option['dynamicWave']):

            # ************************************************************
            # ***** SIDEFLOW
            # ************************************************************

            SideflowChanM3 = self.var.ToChanM3RunoffDt.copy()
            if option['openwaterevapo']:
                SideflowChanM3 -= self.var.EvaAddM3Dt
            if option['wateruse']:
                self.var.WUseAddM3Dt = (self.var.withdrawal_CH_actual_M3_routStep - self.var.returnflow_GwAbs2Channel_M3_routStep)  
                SideflowChanM3 -= self.var.WUseAddM3Dt 
            if option['inflow']:
                SideflowChanM3 += self.var.QInDt
            if option['TransLoss']:
                SideflowChanM3 -= self.var.TransLossM3Dt
            if not(option['InitLisflood']):    # only with no InitLisflood
                if option['simulateLakes']:
                    SideflowChanM3 += self.var.QLakeOutM3Dt                  
                if option['simulateReservoirs']:
                    SideflowChanM3 += self.var.QResOutM3Dt                   
                if option['simulatePolders']:
                    SideflowChanM3 -= self.var.ChannelToPolderM3Dt
                    
                
                    
            ### check mass balance within routing ###
            if option['repMBTs']:  
             if (NoRoutingExecuted<1):
                 self.var.AddedTRUN = np.take(np.bincount(self.var.Catchments, weights=self.var.ToChanM3RunoffDt.copy()),self.var.Catchments)
                 if option['inflow']:
                     self.var.AddedTRUN += np.take(np.bincount(self.var.Catchments, weights=self.var.QInDt),self.var.Catchments)
                 if option['openwaterevapo']:
                     self.var.AddedTRUN -= np.take(np.bincount(self.var.Catchments, weights=self.var.EvaAddM3Dt.copy()),self.var.Catchments)
                 if option['wateruse']:
                     self.var.AddedTRUN -= np.take(np.bincount(self.var.Catchments, weights=self.var.WUseAddM3Dt.copy()),self.var.Catchments)
             else:
                 self.var.AddedTRUN += np.take(np.bincount(self.var.Catchments, weights=self.var.ToChanM3RunoffDt.copy()),self.var.Catchments)
                 if option['inflow']:
                     self.var.AddedTRUN += np.take(np.bincount(self.var.Catchments, weights=self.var.QInDt),self.var.Catchments) 
                 if option['openwaterevapo']:
                     self.var.AddedTRUN -= np.take(np.bincount(self.var.Catchments, weights=self.var.EvaAddM3Dt.copy()),self.var.Catchments)
                 if option['wateruse']:
                     self.var.AddedTRUN -= np.take(np.bincount(self.var.Catchments, weights=self.var.WUseAddM3Dt.copy()),self.var.Catchments)      
                                     
            # Runoff (surface runoff + flow out of Upper and Lower Zone), outflow from
            # reservoirs and lakes and inflow from external hydrographs are added to the channel
            # system (here in [cu m])
            #
            # NOTE: polders currently don't work with kinematic wave, but nevertheless
            # ChannelToPolderM3 is already included in sideflow term (so it's there in case
            # the polder routine is ever modified to make it work with kin. wave)
            # Because of wateruse Sideflow might get even smaller than 0,
            # instead of inflow its outflow

            
            SideflowChan = np.where(self.var.IsChannelKinematic, SideflowChanM3 * self.var.InvChanLength * self.var.InvDtRouting,0)

            # ************************************************************
            # ***** KINEMATIC WAVE                        ****************
            # ************************************************************

            if option['InitLisflood'] or (not(option['SplitRouting'])):
                # if InitLisflood no split routing is use

                #  ---- Single Routing ---------------
                # No split routing
                # side flow consists of runoff (incl. groundwater), inflow from reservoirs (optional) and external inflow hydrographs (optional)
                SideflowChan[np.isnan(SideflowChan)] = 0 # TEMPORARY FIX - SEE DEBUG ABOVE!
 
                self.river_router.kinematicWaveRouting(self.var.ChanQKin, SideflowChan, "main_channel")
                self.var.ChanM3Kin = self.var.ChanLength * self.var.ChannelAlpha * self.var.ChanQKin**self.var.Beta
                # Volume in channel at end of computation step

                self.var.ChanM3Kin = np.maximum(self.var.ChanM3Kin, 0.0)
                # Check for negative volumes at the end of computation step
                self.var.ChanQKin = (self.var.ChanM3Kin * self.var.InvChanLength * self.var.InvChannelAlpha) ** (self.var.InvBeta)
                # Correct negative discharge at the end of computation step

                self.var.ChanQ = self.var.ChanQKin.copy()
                # at single kin. ChanQ is the same
                
                self.var.sumDisDay += self.var.ChanQ
                # Total channel storage [cu m], equal to ChanM3Kin
                
                

            else:

                #  ---- Double Routing ---------------
                # routing is split in two (virtual) channels)

                # Ad
                SideflowRatio = np.where((self.var.ChanM3Kin + self.var.Chan2M3Kin) > 0, self.var.ChanM3Kin/(self.var.ChanM3Kin+self.var.Chan2M3Kin), 0.0)
                

          
                # CM ##################################
                # self.var.Sideflow1Chan = np.where(self.var.ChanM3Kin > self.var.M3Limit, SideflowRatio*SideflowChan, SideflowChan)
                # This is creating instability because ChanM3Kin can be < M3Limit between two routing sub-steps
                # TO BY REPLACED WITH THE FOLLOWING
                self.var.Sideflow1Chan = np.where((self.var.ChanM3Kin + self.var.Chan2M3Kin - self.var.Chan2M3Start) > self.var.M3Limit,
                                                   SideflowRatio*SideflowChan, SideflowChan)
                #######################################
                
                

                self.var.Sideflow1Chan = np.where(np.abs(SideflowChan) < 1e-7, SideflowChan, self.var.Sideflow1Chan)
                # too small values are avoided
                Sideflow2Chan = SideflowChan - self.var.Sideflow1Chan
               
                Sideflow2Chan = Sideflow2Chan + self.var.Chan2QStart * self.var.InvChanLength   # originale
                # as kinematic wave gets slower with less water
                # a constant amount of water has to be added
                # -> add QLimit discharge

                # --- Main channel routing ---
                self.river_router.kinematicWaveRouting(self.var.ChanQKin, self.var.Sideflow1Chan, "main_channel")
                self.var.ChanM3Kin = self.var.ChanLength * self.var.ChannelAlpha * self.var.ChanQKin ** self.var.Beta

                self.var.ChanM3Kin = np.maximum(self.var.ChanM3Kin, 0.0)
                # Check for negative volumes at the end of computation step
                self.var.ChanQKin = (self.var.ChanM3Kin * self.var.InvChanLength * self.var.InvChannelAlpha) ** (self.var.InvBeta)
                # Correct negative discharge at the end of computation step


                # --- Floodplains routing ---
                self.river_router.kinematicWaveRouting(self.var.Chan2QKin, Sideflow2Chan, "floodplains")
                self.var.Chan2M3Kin = self.var.ChanLength * self.var.ChannelAlpha2 * self.var.Chan2QKin ** self.var.Beta

                diffM3 = self.var.Chan2M3Kin - self.var.Chan2M3Start
                self.var.Chan2M3Kin = np.where(diffM3 < 0.0, self.var.Chan2M3Start, self.var.Chan2M3Kin)
                # Check for negative volume in second line of routing at the end of routing substep

                self.var.CrossSection2Area = (self.var.Chan2M3Kin - self.var.Chan2M3Start) * self.var.InvChanLength   
                # Compute cross-section for second line of routing
                
                self.var.Chan2QKin = (self.var.Chan2M3Kin * self.var.InvChanLength * self.var.InvChannelAlpha2) ** (self.var.InvBeta)
                # Correct negative discharge at the end of computation step in second line
              
              
                self.var.ChanQ = np.maximum(self.var.ChanQKin + self.var.Chan2QKin - self.var.QLimit, 0.0)
                
                # Superposition Kinematic
                # Main channel routing and floodplains routing
                #print(np.sum(self.var.sumDisDay))
                self.var.sumDisDay_NOTlast=self.var.sumDisDay.copy()
                self.var.sumDisDay += self.var.ChanQ
                # ----------End splitrouting-------------------------------------------------
                
            ###
            
            # ---- Uncomment lines 603-635 in order to compute the mass balance error within the routing module for the options (i) initial run or (ii) split routing off ----
            '''
            if option['repMBTs']:  
             if option['InitLisflood'] or (not(option['SplitRouting'])):              
                if NoRoutingExecuted == (self.var.NoRoutSteps-1):
                  StorageStep = self.var.ChanM3Kin.copy()               
                  ChanQAvgR = self.var.sumDisDay/self.var.NoRoutSteps
                  sum1=ChanQAvgR.copy()
                  sum1[self.var.AtLastPointC == 0] = 0
                  OutStep = np.take(np.bincount(self.var.Catchments,weights=sum1 * self.var.DtSec),self.var.Catchments) 

                  maskinfo = MaskInfo.instance()
                  DisStructureR = maskinfo.in_zero()
                  DischargeM3StructuresR = maskinfo.in_zero() 
                  if not option['InitLisflood']:
                   if option['simulateReservoirs']:
                     sum1 =self.var.ChanQ.copy()
                     StorageStep =  StorageStep + self.var.ReservoirStorageM3.copy()
                     DisStructureR = np.where(self.var.IsUpsOfStructureKinematicC, sum1 * self.var.DtRouting, 0)
                     DischargeM3StructuresR = np.take(np.bincount(self.var.Catchments, weights=DisStructureR), self.var.Catchments)                
                     DischargeM3StructuresR -= self.var.DischargeM3StructuresIni
                  if not option['InitLisflood']:    
                   if option['simulateLakes']:
                     sum1 =self.var.ChanQ.copy()
                     StorageStep =  StorageStep + self.var.LakeStorageM3Balance.copy()
                     DisStructureR = np.where(self.var.IsUpsOfStructureKinematicC, sum1 * self.var.DtRouting, 0)
                     DischargeM3StructuresR = np.take(np.bincount(self.var.Catchments, weights=DisStructureR), self.var.Catchments)               
                     maskinfo = MaskInfo.instance()
                     DisLake = maskinfo.in_zero()
                     np.put(DisLake, self.var.LakeIndex, 0.5 * self.var.LakeInflowCC * self.var.DtRouting)
                     DischargeM3Lake = np.take(np.bincount(self.var.Catchments, weights=DisLake),self.var.Catchments)
                     DischargeM3StructuresR += DischargeM3Lake
                     DischargeM3StructuresR -= self.var.DischargeM3StructuresIni
                  # Total Mass Balance Error in m3 per catchment for Initial Run OR Split Routing OFF
                  MB =- np.sum(StorageStep)+np.sum(self.var.StorageStepINIT) - OutStep[0]  -DischargeM3StructuresR[0] +self.var.AddedTRUN                     
                  self.var.StorageStepINIT= np.sum(StorageStep) + DischargeM3StructuresR[0]                  
            '''
            if option['repMBTs']:  
             if (not(option['InitLisflood'])) and (option['SplitRouting']):
                # compute the mass balance at the last of the sub-routing steps in order to account for the contributions of lakes and reservoirs
                if NoRoutingExecuted == (self.var.NoRoutSteps-1):
                  ChanQAvgSR = self.var.sumDisDay/self.var.NoRoutSteps
                  sum1=ChanQAvgSR.copy()
                  sum1[self.var.AtLastPointC == 0] = 0
                  OutStep = np.take(np.bincount(self.var.Catchments,weights=sum1 * self.var.DtSec),self.var.Catchments) 
  
                  StorageStep=[]
                  StorageStep= self.var.ChanM3Kin.copy()+self.var.Chan2M3Kin.copy()-self.var.Chan2M3Start.copy()                
                 
                  
                  maskinfo = MaskInfo.instance()
                  DisStructureSR = maskinfo.in_zero()               
                  DischargeM3StructuresR = maskinfo.in_zero() 
                  
                  if option['simulateReservoirs']: 
                     sum1=[]
                     sum1 =self.var.ChanQ.copy()
                     StorageStep =  StorageStep + self.var.ReservoirStorageM3.copy()
                     DisStructureSR = np.where(self.var.IsUpsOfStructureKinematicC, sum1 * self.var.DtRouting, 0)
                     DischargeM3StructuresR = np.take(np.bincount(self.var.Catchments, weights=DisStructureSR), self.var.Catchments)                
                     DischargeM3StructuresR -= self.var.DischargeM3StructuresIni
                       
                  if option['simulateLakes']:
                     sum1 =self.var.ChanQ.copy()
                     StorageStep =  StorageStep + self.var.LakeStorageM3Balance.copy()
                     DisStructureSR = np.where(self.var.IsUpsOfStructureKinematicC, sum1 * self.var.DtRouting, 0)
                     DischargeM3StructuresR = np.take(np.bincount(self.var.Catchments, weights=DisStructureSR), self.var.Catchments)               
                     DisLake = maskinfo.in_zero()
                     np.put(DisLake, self.var.LakeIndex, 0.5 * self.var.LakeInflowCC * self.var.DtRouting)
                     DischargeM3Lake = np.take(np.bincount(self.var.Catchments, weights=DisLake),self.var.Catchments)
                     DischargeM3StructuresR += DischargeM3Lake

                     DischargeM3StructuresR -= self.var.DischargeM3StructuresIni                    
                                        
                  # Mass Balance Error due to the Split Routing module
                  StorageStep1=np.take(np.bincount(self.var.Catchments, weights=StorageStep), self.var.Catchments)
                  
                  self.var.MBErrorSplitRoutingM3  = - StorageStep1 + self.var.StorageStepINIT - OutStep  - DischargeM3StructuresR + self.var.AddedTRUN
                  # Discharge error at the outlet pointt [m3/s]
                  QoutCorrection=self.var.MBErrorSplitRoutingM3/self.var.DtRouting                  
                  QoutCorrection[self.var.AtLastPointC == 0] = 0
                  self.var.OutletDischargeErrorSplitRouting = np.take(np.bincount(self.var.Catchments,weights=QoutCorrection),self.var.Catchments)                  

                  self.var.StorageStepINIT= StorageStep1.copy()+DischargeM3StructuresR 
 


            TotalCrossSectionArea = np.maximum(self.var.ChanM3Kin*self.var.InvChanLength,0.01)

            self.var.FlowVelocity = np.minimum(self.var.ChanQKin/TotalCrossSectionArea, 0.36*self.var.ChanQKin**0.24)
            # Channel velocity (m/s); dividing Q (m3/s) by CrossSectionArea (m2)
            # avoid extreme velocities by using the Wollheim 2006 equation
            # assume 0.1 for upstream areas (outside ChanLdd)
            self.var.FlowVelocity *= np.minimum(np.sqrt(self.var.PixelArea)*self.var.InvChanLength,1);
            # reduction for sinuosity of channels
            self.var.TravelDistance=self.var.FlowVelocity*self.var.DtSec;
            # if flow is fast, Traveltime=1, TravelDistance is high: Pixellength*DtSec
            # if flow is slow, Traveltime=DtSec then TravelDistance=PixelLength
            # maximum set to 30km/day for 5km cell, is at DtSec/Traveltime=6, is at Traveltime<DtSec/6



"""

Copyright 2019 European Union

Licensed under the EUPL, Version 1.2 or as soon they will be approved by the European Commission
subsequent versions of the EUPL (the "Licence");

You may not use this work except in compliance with the Licence.
You may obtain a copy of the Licence at:

https://joinup.ec.europa.eu/sites/default/files/inline-files/EUPL%20v1_2%20EN(1).txt

Unless required by applicable law or agreed to in writing,
software distributed under the Licence is distributed on an "AS IS" basis,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the Licence for the specific language governing permissions and limitations under the Licence.

"""
from __future__ import absolute_import, print_function

import os
import warnings

from ..global_modules.errors import LisfloodWarning
from ..global_modules.settings import LisSettings


def is_number(v):
    if not v:
        return False
    try:
        float(v)
    except ValueError:
        return False
    else:
        return True


def is_path(v):
    if not v:
        return False
    path_no_ext, current_ext = os.path.splitext(v)
    v_alt = '{}.{}'.format(path_no_ext, 'nc' if current_ext in ('.map', '') else 'map')
    a = (os.path.exists(v) and os.path.isfile(v)) or os.access(v, os.W_OK)
    b = (os.path.exists(v_alt) and os.path.isfile(v_alt)) or os.access(v_alt, os.W_OK)
    return a or b


class HydroModule(object):
    # bindings to check for each options used by the HydroModule
    # e.g. in lakes.py for Lakes module
    # input_files_keys = {'simulateLakes': ['LakeSites', 'TabLakeArea', 'TabLakeA', 'LakeMultiplier',
    #                                       'LakeInitialLevelValue', 'TabLakeAvNetInflowEstimate', 'PrevDischarge',
    #                                       'LakePrevInflowValue', 'LakePrevOutflowValue']}
    input_files_keys = None
    module_name = None

    @classmethod
    def check_input_files(cls, option):
        settings = LisSettings.instance()
        binding = settings.binding
        keys = cls.input_files_keys.get(option, [])
        errors = []
        for k in keys:
            msg = None
            k_ok = True
            if not binding.get(k):
                msg = '[{}]: setting "{}" is missing in settings file {}'.format(cls.module_name, k, settings.settings_path)
                k_ok = False
            elif not (is_path(binding[k]) or is_number(binding[k])):
                k_ok = False
                msg = '[{}]: setting {} refers to a non existing path or a not well-formed float value: {}'.format(cls.module_name, k, binding[k])
            if not k_ok:
                warnings.warn(LisfloodWarning(msg))
                errors.append(msg)
        return errors



"""

Copyright 2019 European Union

Licensed under the EUPL, Version 1.2 or as soon they will be approved by the European Commission  subsequent versions of the EUPL (the "Licence");

You may not use this work except in compliance with the Licence.
You may obtain a copy of the Licence at:

https://joinup.ec.europa.eu/sites/default/files/inline-files/EUPL%20v1_2%20EN(1).txt

Unless required by applicable law or agreed to in writing, software distributed under the Licence is distributed on an "AS IS" basis,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the Licence for the specific language governing permissions and limitations under the Licence.

"""

from __future__ import print_function, absolute_import

import uuid
import os
import sys
import warnings

src_dir = os.path.dirname(os.path.abspath(__file__))
if os.path.exists(src_dir):
    sys.path.append(src_dir)

from pcraster.framework import EnsKalmanFilterFramework, MonteCarloFramework

from .global_modules.checkers import ModulesInputs, MeteoForcings
from .global_modules.zusatz import DynamicFramework
from .Lisflood_EnKF import LisfloodModel_EnKF
from .Lisflood_dynamic import LisfloodModel_dyn

from .Lisflood_initial import LisfloodModel_ini
from .Lisflood_monteCarlo import LisfloodModel_monteCarlo
from .global_modules.settings import LisSettings, CDFFlags
from .global_modules.errors import LisfloodError
from lisflood.global_modules.settings import LisfloodRunInfo
from .global_modules.settings import calendar, inttodate
from . import __authors__, __version__, __date__, __status__


class LisfloodModel(LisfloodModel_ini, LisfloodModel_dyn, LisfloodModel_monteCarlo, LisfloodModel_EnKF):
    """ Joining the initial and the dynamic part
        of the Lisflood model
    """
    pass

# ==================================================
# ============== LISFLOOD execute ==================
# ==================================================


def lisfloodexe(lissettings=None):

    _ = CDFFlags(uuid.uuid4())
    if isinstance(lissettings, str):
        # we passed file path
        lissettings = LisSettings(lissettings)
    else:
        # deal with LisSettings instance
        lissettings = LisSettings.instance() if lissettings is None else lissettings

    try:
        ModulesInputs.check(lissettings)
        MeteoForcings.check(lissettings)
    except LisfloodError as e:
        # FIXME using logging instead of print statements
        print(e)
        sys.exit(1)

    binding = lissettings.binding
    option = lissettings.options
    report_steps = lissettings.report_steps
    flags = lissettings.flags
    model_steps = lissettings.model_steps
    filter_steps = lissettings.filter_steps
    # read all the possible option for modelling and for generating output
    # read the settingsfile with all information about the catchments(s)
    # and the choosen option for mdelling and output

    # remove steps from ReportSteps that are not included in simulation period
    for key in report_steps:
        report_steps[key] = [x for x in report_steps[key] if model_steps[0] <= x <= model_steps[1]]

    if not flags['veryquiet']:
        print("\nStart Step - End Step: ", model_steps[0], " - ", model_steps[1])
        print("Start Date - End Date: ",
            inttodate(model_steps[0] - 1, calendar(binding['CalendarDayStart'], binding['calendar_type'])),
            " - ",
            inttodate(model_steps[1] - 1, calendar(binding['CalendarDayStart'], binding['calendar_type'])))

    if flags['loud']:
        # print state file date
        print("State file Date: ")
        try:
            print(inttodate(calendar(binding["timestepInit"], binding['calendar_type']),
                            calendar(binding['CalendarDayStart'], binding['calendar_type'])))
        except:
            print(calendar(binding["timestepInit"], binding['calendar_type']))

        # CM: print start step and end step for reporting model state maps
        print("Start Rep Step  - End Rep Step: ", report_steps['rep'][0], " - ", report_steps['rep'][-1])
        print("Start Rep Date  - End Rep Date: ",
              inttodate(calendar(report_steps['rep'][0] - 1, binding['calendar_type']), calendar(binding['CalendarDayStart'], binding['calendar_type'])),
              " - ",
              inttodate(calendar(report_steps['rep'][-1] - 1), calendar(binding['CalendarDayStart'], binding['calendar_type'])))
        # messages at model start
        print("%-6s %10s %11s\n" % ("Step", "Date", "Discharge"))

    # Lisflood is an instance of the class LisfloodModel
    # LisfloodModel includes 2 methods : initial and dynamic (formulas applied at every timestep)
    Lisflood = LisfloodModel()
    # stLisflood is an instance of the class DynamicFramework

    stLisflood = DynamicFramework(Lisflood, firstTimestep=model_steps[0], lastTimeStep=model_steps[1])
    stLisflood.rquiet = True
    stLisflood.rtrace = False

    if lissettings.mc_set:
        """
        ----------------------------------------------
        MonteCarlo/Ensemble Kalman Filter models
        ----------------------------------------------
        """
        mc_lisflood = MonteCarloFramework(stLisflood, nrSamples=lissettings.ens_members[0])
        if lissettings.nrcores[0] > 1:
            mc_lisflood.setForkSamples(True, nrCPUs=lissettings.nrcores[0])

        if lissettings.enkf_set:
            kf_lisflood = EnsKalmanFilterFramework(mc_lisflood)
            kf_lisflood.setFilterTimesteps(filter_steps)
            # Ensemble Kalman filter
            model_to_run = kf_lisflood
        else:
            # MonteCarlo
            model_to_run = mc_lisflood
    else:
        """
        ----------------------------------------------
        Deterministic run
        ----------------------------------------------
        """
        model_to_run = stLisflood

    # print info about execution
    if not flags['veryquiet']:
        print(LisfloodRunInfo(model_to_run))
    # actual run of the model
    if flags['initonly']:
        print('initonly flag activated... Stopping now before entering time loop.')
    else:
        model_to_run.run()

# ==================================================
# ============== USAGE ==============================
# ==================================================


def usage():
    """ prints some lines describing how to use this program
        which arguments and parameters it accepts, etc
    """

    print('LisfloodPy - Lisflood using pcraster Python framework')
    print('Authors: ', __authors__)
    print('Version: ', __version__)
    print('Date: ', __date__)
    print('Status: ', __status__)
    print("""
    Arguments list:
    settings.xml     settings file

    -q --quiet       output progression given as .
    -v --veryquiet   no output progression is given
    -l --loud        output progression given as time step, date and discharge
    -c --check       input maps and stack maps are checked, output for each input map BUT no model run
    -h --noheader    .tss file have no header and start immediately with the time series
    -d --debug       debug outputs
    -i --initonly    only run initialisation, not the dynamic loop
    """)
    sys.exit(1)


def headerinfo():

    print("LisfloodPy ", __version__, " ", __date__)
    print("""
Water balance and flood simulation model for large catchments\n
(C) Institute for Environment and Sustainability
    Joint Research Centre of the European Commission
    TP122, I-21020 Ispra (Va), Italy\n""")


# ==================================================
# ============== MAIN ==============================
# ==================================================

def main(*args):

    # if arguments are provided, overwrite sys.argv
    # used when calling lisflood from another python script
    if len(args) > 0:
        options = args
    else: 
        options = sys.argv[1:]

    # if arguments are missing display usage info
    if len(options) < 1:
        usage()

    settings_file = options[0]
    sys_args = options[1:]

    # setting.xml file
    lissettings = LisSettings(settings_file, sys_args)
    flags = lissettings.flags
    if not (flags['veryquiet'] or flags['quiet']):
        headerinfo()
    lisfloodexe(lissettings)



"""

Copyright 2019 European Union

Licensed under the EUPL, Version 1.2 or as soon they will be approved by the European Commission  subsequent versions of the EUPL (the "Licence");

You may not use this work except in compliance with the Licence.
You may obtain a copy of the Licence at:

https://joinup.ec.europa.eu/sites/default/files/inline-files/EUPL%20v1_2%20EN(1).txt

Unless required by applicable law or agreed to in writing, software distributed under the Licence is distributed on an "AS IS" basis,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the Licence for the specific language governing permissions and limitations under the Licence.

"""

from __future__ import absolute_import, print_function

import numpy as np

from ..global_modules.settings import MaskInfo, LisSettings
from ..global_modules.add1 import loadmap, defsoil
from . import HydroModule
from collections import OrderedDict
#from global_modules.add1 import *
import xarray as xr


def pressure2SoilMoistureFun(residual_sm, sat_sm, GenuA, GenuN, GenuM):
    """Generate a function to compute soil moisture values corresponding to characteristic pressure head levels [cm].
       IT WON'T BE NEEDED ANYMORE ONCE DATA STRUCTURES WILL ALSO HAVE A "layer" DIMENSION, TO MERGE LAYER-SPECIFIC VARIABLES (1a, 1b, 2)"""
    def Mualem(pressure):
        return residual_sm + (sat_sm - residual_sm) / ((1 + (GenuA * pressure) ** GenuN) ** GenuM)
    return Mualem


class soil(HydroModule):

    """
    # ************************************************************
    # ***** SOIL         *****************************************
    # ************************************************************
    """
    input_files_keys = {'all': ['SoilDepth1', 'SoilDepth1Forest', 'SoilDepth2', 'SoilDepth2Forest',
                                'SoilDepth3', 'SoilDepth3Forest', 'CourantCrit', 'LeafDrainageTimeConstant',
                                'AvWaterRateThreshold', 'MapCropCoef', 'MapForestCropCoef', 'MapIrrigationCropCoef',
                                'MapCropGroupNumber', 'MapForestCropGroupNumber', 'MapIrrigationCropGroupNumber',
                                'MapN', 'MapForestN', 'MapKSat1', 'MapKSat1Forest', 'MapKSat2', 'MapKSat2Forest', 'MapKSat3',
                                'MapLambda1', 'MapLambda1Forest', 'MapLambda2', 'MapLambda2Forest', 'MapLambda3',
                                'MapGenuAlpha1', 'MapGenuAlpha1Forest', 'MapGenuAlpha2', 'MapGenuAlpha2Forest', 'MapGenuAlpha3',
                                'MapThetaSat1', 'MapThetaSat1Forest', 'MapThetaSat2', 'MapThetaSat2Forest', 'MapThetaSat3',
                                'MapThetaRes1', 'MapThetaRes1Forest', 'MapThetaRes2', 'MapThetaRes2Forest', 'MapThetaRes3',
                                'ThetaInit1Value', 'ThetaForestInit1Value', 'ThetaIrrigationInit1Value',
                                'ThetaInit2Value', 'ThetaForestInit2Value', 'ThetaIrrigationInit2Value',
                                'ThetaInit3Value', 'ThetaForestInit3Value', 'ThetaIrrigationInit3Value',
                                'b_Xinanjiang', 'PowerPrefFlow',
                                'DSLRInitValue', 'DSLRForestInitValue', 'DSLRIrrigationInitValue',
                                'CumIntInitValue', 'CumIntForestInitValue', 'CumIntIrrigationInitValue',
                                'CumIntSealedInitValue', 'SMaxSealed'],
                        'drainedIrrigation': ['DrainedFraction'],
                        'simulatePF': ['HeadMax']}
    module_name = 'Soil'

    def __init__(self, soil_variable):
        self.var = soil_variable

# --------------------------------------------------------------------------
# --------------------------------------------------------------------------

    def initial(self):
        """ initial part of the soil module
        """
        def splitlanduse(array1, array2=None, array3=None):
            """ splits maps into the 3 different land use types - other , forest, irrigation
            """
            if array2 is None:
                array2 = array1
            if array3 is None:
                array3 = array1
            return [array1, array2, array3]

        # ------------------------------------------------
        # Fraction of Landuse
        # using maps instead of table to be more sensitive to landuse change
        # Soil properties for percentage of pixel area without forest, water ,
        # impervious soil
        settings = LisSettings.instance()
        option = settings.options
        binding = settings.binding
        maskinfo = MaskInfo.instance()
        if not option["cropsEPIC"]: # If EPIC is active, the rice fraction initialisation is handled by EPIC (setSoilFractions in EPIC_main.py)
            self.var.SoilFraction.values[self.var.vegetation.index('Rainfed_prescribed')] += self.var.RiceFraction

        # for the moment rice is treated as other fraction
        # if the fraction of water varies then the other fraction are stored
        self.var.WaterFractionBase = self.var.WaterFraction.copy()
        self.var.OtherFractionBase = self.var.OtherFraction.copy()
        self.var.IrrigationFractionBase = self.var.IrrigationFraction.copy()
        self.var.ForestFractionBase = self.var.ForestFraction.copy()
        self.var.DirectRunoffFractionBase = self.var.DirectRunoffFraction.copy()
        self.var.PermeableFraction = 1 - self.var.DirectRunoffFraction - self.var.WaterFraction
        # Permeable fraction of pixel

        # Soil Depth 1st layer [mm]
        # soil depth top layer for every landuse but forest, impervious soil
        # and water

        self.var.SoilDepth1a = self.var.defsoil('SoilDepth1', 'SoilDepth1Forest')
        self.var.SoilDepth1b = self.var.defsoil('SoilDepth2', 'SoilDepth2Forest')
        self.var.SoilDepth2 = self.var.defsoil('SoilDepth3', 'SoilDepth3Forest')
        self.var.SoilDepthTotal = self.var.SoilDepth1a + self.var.SoilDepth1b + self.var.SoilDepth2
        # ----------------- miscParameters ---------------------------

        self.var.CourantCrit = loadmap('CourantCrit')
        self.var.LeafDrainageK = np.minimum(self.var.DtDay * (1 / loadmap('LeafDrainageTimeConstant')), 1)
        # Reservoir constant for interception store
        # Minimize statement ONLY needed if LeafDrainageTimeConstant
        # becomes greater than DtDay (unlikely, but possible for time steps
        # greater than daily)

        self.var.AvWaterThreshold = loadmap('AvWaterRateThreshold') * self.var.DtDay
        # Converts AvWaterRateThreshold from  [mm/day]
        # to [mm] per time step


        # ************************************************************
        # ***** LAND USE RELATED MAPS ********************************
        # ************************************************************

        # using maps instead of table to be more sensitive to landuse change
        # Soil properties for percentage of pixel area without forest, water ,
        # impervious soil
        self.var.CropCoef = self.var.defsoil('MapCropCoef', 'MapForestCropCoef', 'MapIrrigationCropCoef')

        # crop coefficients for each land use class
        self.var.CropGroupNumber = self.var.defsoil('MapCropGroupNumber', 'MapForestCropGroupNumber', 'MapIrrigationCropGroupNumber')

        # crop group numbers for soil water depletion factor        
        self.var.NManning = self.var.defsoil('MapN','MapForestN', 0.02, OrderedDict([self.var.dim_runoff, self.var.dim_pixel]))
        # Manning's roughness coefficient for each land use class
        # third manning is from direct runoff
        # self.var.NManningDirect=scalar(0.02)

        # ************************************************************
        # ***** MAPS DERIVED FROM SOIL TEXTURE AND SOIL DEPTH ********
        # ************************************************************

        # HYPRES Parameters for TopSoil:

        self.var.KSat1a = self.var.defsoil('MapKSat1', 'MapKSat1Forest')
        self.var.KSat1b = self.var.defsoil('MapKSat2', 'MapKSat2Forest')
        self.var.KSat2 = self.var.defsoil('MapKSat3')
        # Saturated conductivity for other land use and forest
        Lambda1a = self.var.defsoil('MapLambda1', 'MapLambda1Forest')
        Lambda1b = self.var.defsoil('MapLambda2', 'MapLambda2Forest')
        Lambda2 = self.var.defsoil('MapLambda3')
        # Pore-size index (for Van Genuchten parameters)
        GenuAlpha1a = self.var.defsoil('MapGenuAlpha1', 'MapGenuAlpha1Forest')
        GenuAlpha1b = self.var.defsoil('MapGenuAlpha2', 'MapGenuAlpha2Forest')
        GenuAlpha2 = self.var.defsoil('MapGenuAlpha3')
        # Van Genuchten Alpha coefficient
        ThetaS1a = self.var.defsoil('MapThetaSat1','MapThetaSat1Forest')
        ThetaS1b = self.var.defsoil('MapThetaSat2','MapThetaSat2Forest')
        ThetaS2 = self.var.defsoil('MapThetaSat3')
        # Soil moisture content at saturation [V/V]
        ThetaRes1a = self.var.defsoil('MapThetaRes1', 'MapThetaRes1Forest')
        ThetaRes1b = self.var.defsoil('MapThetaRes2', 'MapThetaRes2Forest')
        ThetaRes2 = self.var.defsoil('MapThetaRes3')
        # Soil moisture content at saturation [V/V]

        GenuN1a = 1 + Lambda1a
        GenuN1b = 1 + Lambda1b
        GenuN2 = 1 + Lambda2
        # self.var.GenuM1=Lambda1/GenuN1
        self.var.GenuM1a = Lambda1a / GenuN1a
        self.var.GenuM1b = Lambda1b / GenuN1b
        self.var.GenuM2 = Lambda2 / GenuN2
        # self.var.GenuInvM1=1/self.var.GenuM1
        self.var.GenuInvM1a = 1 / self.var.GenuM1a
        self.var.GenuInvM1b = 1 / self.var.GenuM1b
        self.var.GenuInvM2 = 1 / self.var.GenuM2
        # self.var.GenuInvN1=1/GenuN1
        self.var.GenuInvN1a = 1 / GenuN1a
        self.var.GenuInvN1b = 1 / GenuN1b
        self.var.GenuInvN2 = 1 / GenuN2
        # GenuInvAlpha1=1/GenuAlpha1
        self.var.GenuInvAlpha1a = 1 / GenuAlpha1a
        self.var.GenuInvAlpha1b = 1 / GenuAlpha1b
        self.var.GenuInvAlpha2 = 1 / GenuAlpha2
        
        # Soil moisture [mm] at saturation:
        self.var.WS1a = ThetaS1a * self.var.SoilDepth1a
        self.var.WS1b = ThetaS1b * self.var.SoilDepth1b
        self.var.WS2 = ThetaS2 * self.var.SoilDepth2
        self.var.WS1 = self.var.WS1a + self.var.WS1b
        # Residual soil moisture [mm]:
        self.var.WRes1a = ThetaRes1a * self.var.SoilDepth1a
        self.var.WRes1b = ThetaRes1b * self.var.SoilDepth1b
        self.var.WRes2 = ThetaRes2 * self.var.SoilDepth2
        self.var.WRes1 = self.var.WRes1a + self.var.WRes1b
        # Saturation minus residual soil moisture [mm]:
        self.var.WS1WResa = self.var.WS1a - self.var.WRes1a
        self.var.WS1WResb = self.var.WS1b - self.var.WRes1b
        self.var.WS2WRes = self.var.WS2 - self.var.WRes2
        
        # Generate Mualem functions (van Genuchten, 1980) for the 3 soil layers:
        MualemEq1a = pressure2SoilMoistureFun(self.var.WRes1a, self.var.WS1a, GenuAlpha1a, GenuN1a, self.var.GenuM1a)
        MualemEq1b = pressure2SoilMoistureFun(self.var.WRes1b, self.var.WS1b, GenuAlpha1b, GenuN1b, self.var.GenuM1b)
        MualemEq2 = pressure2SoilMoistureFun(self.var.WRes2, self.var.WS2, GenuAlpha2, GenuN2, self.var.GenuM2)
        # Soil moisture at field capacity (pF2, 100 cm) [mm water slice]:       
        self.var.WFC1a = MualemEq1a(100)
        self.var.WFC1b = MualemEq1b(100)
        self.var.WFC2 = MualemEq2(100)
        self.var.WFC1 = self.var.WFC1a + self.var.WFC1b
        # Soil moisture [mm] at pF3, 1000 cm (pressure):
        self.var.WPF3a = MualemEq1a(1000)
        self.var.WPF3b = MualemEq1b(1000)
        self.var.WPF3 = self.var.WPF3a + self.var.WPF3b
        # Soil moisture at wilting point (pF4.2, 10**4.2 cm) in [mm] water slice:
        self.var.WWP1a = MualemEq1a(15000)
        self.var.WWP1b = MualemEq1b(15000)
        self.var.WWP2 = MualemEq2(15000)
        self.var.WWP1 = self.var.WWP1a + self.var.WWP1b       
       
        self.var.PoreSpaceNotZero1a = np.logical_and(self.var.SoilDepth1a != 0, self.var.WS1a != 0)
        self.var.PoreSpaceNotZero1b = np.logical_and(self.var.SoilDepth1b != 0, self.var.WS1b != 0)
        self.var.PoreSpaceNotZero2 = np.logical_and(self.var.SoilDepth2 != 0, self.var.WS2 != 0)

        # ***** INITIAL VALUES for soil
        
        if not option["cropsEPIC"]: 
       
           ThetaInit1aValue = self.var.allocateVariableAllVegetation()
           ThetaInit1bValue = self.var.allocateVariableAllVegetation()
           ThetaInit2Value = self.var.allocateVariableAllVegetation()

           iveg = self.var.vegetation.index('Rainfed_prescribed')           
           ThetaInit1aValue.values[iveg] = loadmap('ThetaInit1Value')
           ThetaInit1bValue.values[iveg] = loadmap('ThetaInit2Value')
           ThetaInit2Value.values[iveg] = loadmap('ThetaInit3Value')
           
           iveg = self.var.vegetation.index('Forest_prescribed')
           ThetaInit1aValue.values[iveg] = loadmap('ThetaForestInit1Value')
           ThetaInit1bValue.values[iveg] = loadmap('ThetaForestInit2Value')
           ThetaInit2Value.values[iveg] = loadmap('ThetaForestInit3Value')
           
           iveg = self.var.vegetation.index('Irrigated_prescribed')
           ThetaInit1aValue.values[iveg] = loadmap('ThetaIrrigationInit1Value')
           ThetaInit1bValue.values[iveg] = loadmap('ThetaIrrigationInit2Value')
           ThetaInit2Value.values[iveg] = loadmap('ThetaIrrigationInit3Value')
        else:
            ThetaInit1aValue = self.var.initialiseVariableAllVegetation('ThetaInit1Value')
            ThetaInit1bValue = self.var.initialiseVariableAllVegetation('ThetaInit2Value')
            ThetaInit2Value = self.var.initialiseVariableAllVegetation('ThetaInit3Value')
        
  
        # Allocate soil moisture [mm] data structures (merge them into a single xarray.DataArray by adding a "layer" dimension !!!)
        self.var.W1a = self.var.allocateVariableAllVegetation()
        self.var.W1b = self.var.allocateVariableAllVegetation()
        self.var.W2 = self.var.allocateVariableAllVegetation()
        # Initialise soil moisture in permeable fraction of pixel:
        # Set soil to field capacity if initial value in settings file equals -9999;
        # Set to zero if soil depth is zero.
        # IMPORTANT: WInit1 and WInit2 represent the soil moisture in the *permeable* fraction of the pixel *only*
        # (since all soil moisture-related calculations are done for permeable fraction only!).

        for veg, luse in self.var.VEGETATION_LANDUSE.items():
            iveg = self.var.vegetation.index(veg)
            iluse = self.var.SOIL_USES.index(luse)
            ini_1a = np.where(ThetaInit1aValue[iveg] == -9999, self.var.WFC1a[iluse], ThetaInit1aValue[iveg] * self.var.SoilDepth1a[iluse])
            self.var.W1a[iveg] = np.where(self.var.PoreSpaceNotZero1a[iluse], ini_1a, 0)
            ini_1b = np.where(ThetaInit1bValue[iveg] == -9999, self.var.WFC1b[iluse], ThetaInit1bValue[iveg] * self.var.SoilDepth1b[iluse])
            self.var.W1b[iveg] = np.where(self.var.PoreSpaceNotZero1b[iluse], ini_1b, 0)
            ini_2 = np.where(ThetaInit2Value[iveg] == -9999, self.var.WFC2[iluse], ThetaInit2Value[iveg] * self.var.SoilDepth2[iluse])
            self.var.W2[iveg] = np.where(self.var.PoreSpaceNotZero2[iluse], ini_2, 0)
        self.var.W1 = self.var.W1a + self.var.W1b
                    

        self.var.Sat1a = self.var.allocateVariableAllVegetation()
        self.var.Sat1b = self.var.allocateVariableAllVegetation()
        self.var.Sat1 = self.var.allocateVariableAllVegetation()
        self.var.Sat2 = self.var.allocateVariableAllVegetation()
        '''
        self.var.PoreSpaceNotZero1a = splitlanduse(np.bool8((self.var.SoilDepth1a[0] != 0) & (self.var.WS1a[0] != 0)),
                                                   np.bool8((self.var.SoilDepth1a[1] != 0) & (self.var.WS1a[1] != 0)),
                                                   np.bool8((self.var.SoilDepth1a[2] != 0) & (self.var.WS1a[2] != 0)))
        self.var.PoreSpaceNotZero1b = splitlanduse(np.bool8((self.var.SoilDepth1b[0] != 0) & (self.var.WS1b[0] != 0)),
                                                   np.bool8((self.var.SoilDepth1b[1] != 0) & (self.var.WS1b[1] != 0)),
                                                   np.bool8((self.var.SoilDepth1b[2] != 0) & (self.var.WS1b[2] != 0)))
        self.var.PoreSpaceNotZero2 = splitlanduse(np.bool8((self.var.SoilDepth2[0] != 0) & (self.var.WS2[0] != 0)),
                                                  np.bool8((self.var.SoilDepth2[1] != 0) & (self.var.WS2[1] != 0)),
                                                  np.bool8((self.var.SoilDepth2[2] != 0) & (self.var.WS2[2] != 0)))

#        self.var.PoreSpaceNotZero1 = [np.bool8((self.var.SoilDepth1[0] != 0) & (self.var.WS1[0] != 0)),
#            np.bool8((self.var.SoilDepth1[1] != 0) & (self.var.WS1[1] != 0))]
#        self.var.PoreSpaceNotZero2 = [np.bool8((self.var.SoilDepth2[0] != 0) & (self.var.WS2[0] != 0)),
#            np.bool8((self.var.SoilDepth2[1] != 0) & (self.var.WS2[1] != 0))]
        # Flag that is Boolean 1 if pixel has (any) soil moisture storage space.
        # Needed to avoid divisions by zero in soil moisture calculations

        # ***** INITIAL VALUES for soil

        ThetaInit1aValue = defsoil('ThetaInit1Value', 'ThetaForestInit1Value', 'ThetaIrrigationInit1Value')
        ThetaInit1bValue = defsoil('ThetaInit2Value', 'ThetaForestInit2Value', 'ThetaIrrigationInit2Value')
        ThetaInit2Value = defsoil('ThetaInit3Value', 'ThetaForestInit3Value', 'ThetaIrrigationInit3Value')

        WInit1a = splitlanduse(maskinfo.in_zero())
        WInit1b = splitlanduse(maskinfo.in_zero())
        WInit2 = splitlanduse(maskinfo.in_zero())
        WInit1a[0] = np.where(ThetaInit1aValue[0] == -9999, self.var.WFC1a[0], ThetaInit1aValue[0] * self.var.SoilDepth1a[0])
        WInit1a[1] = np.where(ThetaInit1aValue[1] == -9999, self.var.WFC1a[1], ThetaInit1aValue[1] * self.var.SoilDepth1a[1])
        WInit1a[2] = np.where(ThetaInit1aValue[2] == -9999, self.var.WFC1a[2], ThetaInit1aValue[2] * self.var.SoilDepth1a[2])
        WInit1b[0] = np.where(ThetaInit1bValue[0] == -9999, self.var.WFC1b[0], ThetaInit1bValue[0] * self.var.SoilDepth1b[0])
        WInit1b[1] = np.where(ThetaInit1bValue[1] == -9999, self.var.WFC1b[1], ThetaInit1bValue[1] * self.var.SoilDepth1b[1])
        WInit1b[2] = np.where(ThetaInit1bValue[2] == -9999, self.var.WFC1b[2], ThetaInit1bValue[2] * self.var.SoilDepth1b[2])
        WInit2[0] = np.where(ThetaInit2Value[0] == -9999, self.var.WFC2[0], ThetaInit2Value[0] * self.var.SoilDepth2[0])
        WInit2[1] = np.where(ThetaInit2Value[1] == -9999, self.var.WFC2[1], ThetaInit2Value[1] * self.var.SoilDepth2[1])
        WInit2[2] = np.where(ThetaInit2Value[2] == -9999, self.var.WFC2[2], ThetaInit2Value[2] * self.var.SoilDepth2[2])
        # Initial soil moisture in permeable fraction of pixel
        # Set soil to field capacity if initial value
        # in settings file equals -9999
        # IMPORTANT: WInit1 and WInit2 represent the soil moisture in the
        # *permeable* fraction of the pixel *only* (since all soil moisture-related
        # calculations are done for permeable fraction only!).

        WInit1a[0] = np.where(self.var.PoreSpaceNotZero1a[0], WInit1a[0], maskinfo.in_zero())
        WInit1a[1] = np.where(self.var.PoreSpaceNotZero1a[1], WInit1a[1], maskinfo.in_zero())
        WInit1a[2] = np.where(self.var.PoreSpaceNotZero1a[2], WInit1a[2], maskinfo.in_zero())
        WInit1b[0] = np.where(self.var.PoreSpaceNotZero1b[0], WInit1b[0], maskinfo.in_zero())
        WInit1b[1] = np.where(self.var.PoreSpaceNotZero1b[1], WInit1b[1], maskinfo.in_zero())
        WInit1b[2] = np.where(self.var.PoreSpaceNotZero1b[2], WInit1b[2], maskinfo.in_zero())
        WInit2[0] = np.where(self.var.PoreSpaceNotZero2[0], WInit2[0], maskinfo.in_zero())
        WInit2[1] = np.where(self.var.PoreSpaceNotZero2[1], WInit2[1], maskinfo.in_zero())
        WInit2[2] = np.where(self.var.PoreSpaceNotZero2[2], WInit2[2], maskinfo.in_zero())
        # Set to zero if soil depth is zero
        self.var.W1a = WInit1a
        self.var.W1b = WInit1b
        self.var.W1 = np.add(self.var.W1a, self.var.W1b)
        self.var.W2 = WInit2
        # Set soil moisture to initial value

        self.var.Sat1a = splitlanduse(maskinfo.in_zero())
        self.var.Sat1b = splitlanduse(maskinfo.in_zero())
        self.var.Sat1 = splitlanduse(maskinfo.in_zero())
        self.var.Sat2 = splitlanduse(maskinfo.in_zero())
        '''

        # ************************************************************
        # ***** REPEATEDLY USED EXPRESSIONS IN XINANJIANG
        # ***** INFILTRATION MODEL
        # ************************************************************
        self.var.b_Xinanjiang = loadmap('b_Xinanjiang')
    
        if isinstance(self.var.b_Xinanjiang, float):
           MAP = []
           MAP = maskinfo.in_zero() +   self.var.b_Xinanjiang
           self.var.b_Xinanjiang = []
           self.var.b_Xinanjiang = MAP
               
        self.var.PowerInfPot = (self.var.b_Xinanjiang + 1) / self.var.b_Xinanjiang
        # Power in infiltration equation
        self.var.StoreMaxPervious = self.var.WS1 / (self.var.b_Xinanjiang + 1)
        # Maximum soil moisture storage in pervious fraction of
        # pixel (expressed as [mm] water slice)

        # ********************************************
        # ****  PowerPrefFlow
        # ********************************************
        self.var.PowerPrefFlow = loadmap('PowerPrefFlow')
                
        if isinstance(self.var.PowerPrefFlow, float):
           MAP = []
           MAP = maskinfo.in_zero() +   self.var.PowerPrefFlow
           self.var.PowerPrefFlow = []
           self.var.PowerPrefFlow = MAP

           
        # ************************************************************
        # ***** INITIAL VALUES
        # ************************************************************
        # Inputs in waterbalance model and/or initial assumption
        # CMmod
        # DSLRInit = defsoil('DSLRInitValue', 'DSLRForestInitValue','DSLRIrrigationInitValue')
        
        
        if not option["cropsEPIC"]: 
           self.var.DSLR = self.var.allocateVariableAllVegetation()
           ivegRainfedPrescribed = self.var.vegetation.index('Rainfed_prescribed')
           ivegForestPrescribed = self.var.vegetation.index('Forest_prescribed')
           ivegIrrigatedPrescribed = self.var.vegetation.index('Irrigated_prescribed')

           self.var.DSLR[ivegRainfedPrescribed] = loadmap('DSLRInitValue')
           self.var.DSLR[ivegForestPrescribed]  = loadmap('DSLRForestInitValue')
           self.var.DSLR[ivegIrrigatedPrescribed]  = loadmap('DSLRIrrigationInitValue')
           self.var.DSLR[ivegRainfedPrescribed]  = np.where(self.var.DSLR[ivegRainfedPrescribed]<1,1,self.var.DSLR[ivegRainfedPrescribed])
           self.var.DSLR[ivegForestPrescribed]  = np.where(self.var.DSLR[ivegForestPrescribed]<1,1,self.var.DSLR[ivegForestPrescribed])
           self.var.DSLR[ivegIrrigatedPrescribed]  = np.where(self.var.DSLR[ivegIrrigatedPrescribed]<1,1,self.var.DSLR[ivegIrrigatedPrescribed])
        else:
           self.var.DSLR = np.maximum(self.var.initialiseVariableAllVegetation('DSLRInitValue'), 1) # DSLR<1 may cause square root of negative number, if Dtsec > 86400!
        
        if not option["cropsEPIC"]:  
           self.var.CumInterception = self.var.allocateVariableAllVegetation()
           self.var.CumInterception[ivegRainfedPrescribed] = loadmap('CumIntInitValue')
           self.var.CumInterception[ivegForestPrescribed] = loadmap('CumIntForestInitValue')
           self.var.CumInterception[ivegIrrigatedPrescribed] = loadmap('CumIntIrrigationInitValue')
        else:
           self.var.CumInterception = self.var.initialiseVariableAllVegetation('CumIntInitValue')
        
        # Initialising cumulative output variables  needed to compute the cumulative mass balance error
        self.var.TotalPrecipitation =  maskinfo.in_zero() # precipitation [mm]
        self.var.TaCUM =  maskinfo.in_zero() # Cumulative transpiration [mm]
        self.var.TaWB = maskinfo.in_zero() # Cumulative transpiration [mm] at the end of the computational time step (waterbalance.py)
        self.var.TaInterceptionCUM =  maskinfo.in_zero() # Cumulative evaporation from interception store [mm]
        self.var.TaInterceptionWB =  maskinfo.in_zero() # Evaporation from interception store [mm] at the end of the computational time step (waterbalance.py)
        self.var.ESActCUM =  maskinfo.in_zero() # Cumulative evaporation [mm]
        self.var.ESActWB = maskinfo.in_zero() # Cumulative evaporation [mm] at the end of the computational time step (waterbalance.py)
        self.var.SoilMoistureStressDays = self.var.allocateVariableAllVegetation() # number of days in simulation with soil moisture stress (days)
        self.var.Theta1a = self.var.allocateVariableAllVegetation() # Theta values are just allocated here - their values are dynamically compute in soilloop.py
        self.var.Theta1b = self.var.allocateVariableAllVegetation()
        self.var.Theta2 = self.var.allocateVariableAllVegetation()
        self.var.TaInterception = self.var.allocateVariableAllVegetation()
        self.var.Interception = self.var.allocateVariableAllVegetation()
        self.var.LeafDrainage = self.var.allocateVariableAllVegetation()
        self.var.potential_transpiration = self.var.allocateVariableAllVegetation()
        self.var.Ta = self.var.allocateVariableAllVegetation()
        self.var.ESAct = self.var.allocateVariableAllVegetation()
        self.var.PrefFlow = self.var.allocateVariableAllVegetation()
        self.var.Infiltration = self.var.allocateVariableAllVegetation()
        self.var.SeepTopToSubA = self.var.allocateVariableAllVegetation()
        self.var.SeepTopToSubB = self.var.allocateVariableAllVegetation()
        self.var.SeepSubToGW = self.var.allocateVariableAllVegetation()
        self.var.Theta = self.var.allocateVariableAllVegetation()
        self.var.AvailableWaterForInfiltration = self.var.allocateVariableAllVegetation()
        ivegvalues = []
        for veg in self.var.prescribed_vegetation:
            ivegvalues.append(self.var.vegetation.index(veg))
        self.var.RWS = self.var.allocateVariableAllVegetation()[ivegvalues]


        # ************************************************************
        # ***** INITIALIZATION OF IMPERVIOUS SOIL     ****************
        # ************************************************************

        self.var.CumInterSealed = loadmap('CumIntSealedInitValue')
        
        # cumulative depression storage
        self.var.SMaxSealed = loadmap('SMaxSealed')
        # maximum depression storage for water on impervious surface
        # which is not immediately causing surface runoff  [mm]

        settings = LisSettings.instance()
        option = settings.options

        if option['drainedIrrigation']:
            self.var.DrainedFraction = loadmap('DrainedFraction')
        else: 
            self.var.DrainedFraction = 0.0 ##### STEF: the master simply do not assign this!
            # drained fraction: water from irrigated water is directly send to channel
            # depending on the drained fraction

# ************************************************************
# ***** INITIALIZATION OF PF calculation      ****************
# ************************************************************
        if option['simulatePF']:
            self.var.HeadMax = loadmap('HeadMax')
            self.var.pF0 = self.var.allocateVariableAllVegetation()
            self.var.pF1 = self.var.allocateVariableAllVegetation()
            self.var.pF2 = self.var.allocateVariableAllVegetation()

    def dynamic_perpixel(self):
        """ dynamic part of the soil module
            Calculation per Pixel
        """
        self.var.TaInterceptionAll = self.var.deffraction(self.var.TaInterception) + self.var.DirectRunoffFraction * self.var.TASealed
        self.var.TaInterceptionCUM += self.var.TaInterceptionAll
        self.var.TaInterceptionWB = self.var.TaInterceptionAll
        # Cumulative evaporation of intercepted water [mm]

        self.var.TaPixel = self.var.deffraction(self.var.Ta)
        # pixel-average transpiration in [mm] per timestep
        # (no transpiration from direct runoff fraction)
        self.var.TaCUM += self.var.TaPixel
        self.var.TaWB = self.var.TaPixel
        self.var.ESActPixel = self.var.deffraction(self.var.ESAct) + self.var.WaterFraction * self.var.EWaterAct
        # Pixel-average soil evaporation in [mm] per time step
        # (no evaporation from direct runoff fraction)
        self.var.ESActCUM += self.var.ESActPixel
        self.var.ESActWB = self.var.ESActPixel

        self.var.PrefFlowPixel = self.var.deffraction(self.var.PrefFlow)
        # Pixel-average preferential flow in [mm] per timestep
        # (no preferential flow from direct runoff fraction)
        self.var.InfiltrationPixel = self.var.deffraction(self.var.Infiltration)
        # Pixel-average infiltration in [mm] per timestep
        # (no infiltration in direct runoff fraction)
        tot_sm = self.var.W1a + self.var.W1b + self.var.W2
        for landuse, veg_list in self.var.LANDUSE_VEGETATION.items():
            iveg_list,iveg_list_pres,ilanduse = self.var.get_indexes_from_landuse_and_veg_list_GLOBAL(landuse, veg_list)
            self.var.Theta[iveg_list] = self.var.SoilFraction[iveg_list_pres] * tot_sm[iveg_list] / self.var.SoilDepthTotal[ilanduse]
        ax_veg=self.var.SoilFraction.dims.index("vegetation")
        soil_fract_sum = np.sum(self.var.SoilFraction,ax_veg)
        self.var.ThetaAll = np.where(soil_fract_sum > 0, np.sum(self.var.Theta,ax_veg) / soil_fract_sum, 0)

        self.var.SeepTopToSubPixelA = self.var.deffraction(self.var.SeepTopToSubA)
        self.var.SeepTopToSubPixelB = self.var.deffraction(self.var.SeepTopToSubB)
        self.var.SeepSubToGWPixel = self.var.deffraction(self.var.SeepSubToGW)
        # Pixel-average seepage values in [mm] per timestep
        # (no seepage from direct runoff fraction)
        
        # the variables below were added to report catchment-averaged soil moisture profiles
        self.var.Theta1aPixel = self.var.deffraction(self.var.Theta1a)
        self.var.Theta1bPixel = self.var.deffraction(self.var.Theta1b)
        self.var.Theta2Pixel = self.var.deffraction(self.var.Theta2)



"""
Copyright 2019 European Union
Licensed under the EUPL, Version 1.2 or as soon they will be approved by the European Commission  subsequent versions of the EUPL (the "Licence");
You may not use this work except in compliance with the Licence.
You may obtain a copy of the Licence at:
https://joinup.ec.europa.eu/sites/default/files/inline-files/EUPL%20v1_2%20EN(1).txt
Unless required by applicable law or agreed to in writing, software distributed under the Licence is distributed on an "AS IS" basis,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the Licence for the specific language governing permissions and limitations under the Licence.
"""

from __future__ import absolute_import, print_function

from ..global_modules.add1 import loadmap, readnetcdf
from ..global_modules.settings import LisSettings, EPICSettings
from . import HydroModule

from ..global_modules.add1 import NumpyModified

from collections import OrderedDict
import pandas as pd
import xarray as xr
import numpy as np

class landusechange(HydroModule):

    """
    # ************************************************************
    # ***** LAND USE CHANGE : FRACTION MAPS **********************
    # ************************************************************

    # Each pixel is divided into several fractions, adding up to 1
    # open water
    # forest
    # sealed fraction
    # irrigated areas
    # rice irrigation areas
    # other
    """
    input_files_keys = {'all': ['ForestFraction', 'DirectRunoffFraction', 'WaterFraction',
                                'IrrigationFraction', 'RiceFraction', 'OtherFraction'],
                        'TransientLandUseChange': ['ForestFractionMaps', 'DirectRunoffFractionMaps',
                                                   'WaterFractionMaps',
                                                   'IrrigationFractionMaps', 'RiceFractionMaps', 'OtherFractionMaps']}
    module_name = 'LandUseChange'

    def __init__(self, landusechange_variable):
        self.var = landusechange_variable

# --------------------------------------------------------------------------
# --------------------------------------------------------------------------

    def initial(self):
    
        settings = LisSettings.instance()
        option = settings.options
        binding = settings.binding
        
        """ initial part of the landusechange module
        """
        self.var.ForestFraction = loadmap('ForestFraction', timestampflag='closest').copy()
        self.var.DirectRunoffFraction = loadmap('DirectRunoffFraction', timestampflag='closest').copy()
        self.var.WaterFraction = loadmap('WaterFraction', timestampflag='closest').copy()
        self.var.IrrigationFraction = loadmap('IrrigationFraction', timestampflag='closest').copy()
        self.var.RiceFraction = loadmap('RiceFraction', timestampflag='closest').copy()
        self.var.OtherFraction = loadmap('OtherFraction', timestampflag='closest').copy()
        ##self.var.OtherFraction = 1 - (self.var.RiceFraction+self.var.IrrigationFraction+self.var.WaterFraction+self.var.DirectRunoffFraction+self.var.ForestFraction) 
        ## uncomment the line above to check the impact of pixels where the sum is not 1.0 (NB: this is an error in the fraction input maps as the fractions must always add up to 1.0)
        
        if option['TransientLandUseChange'] and option['readNetcdfStack']:
                model_steps = settings.model_steps                
                self.var.ForestFraction = readnetcdf(binding['ForestFractionMaps'], model_steps[0] , timestampflag='closest')        
                self.var.DirectRunoffFraction = readnetcdf(binding['DirectRunoffFractionMaps'],  model_steps[0] , timestampflag='closest')
                self.var.WaterFraction = readnetcdf(binding['WaterFractionMaps'],  model_steps[0], timestampflag='closest')
                self.var.IrrigationFraction = readnetcdf(binding['IrrigationFractionMaps'],  model_steps[0], timestampflag='closest')
                self.var.RiceFraction = readnetcdf(binding['RiceFractionMaps'],  model_steps[0] , timestampflag='closest')
                self.var.OtherFraction = readnetcdf(binding['OtherFractionMaps'],  model_steps[0] , timestampflag='closest')     
                
        epic_settings = EPICSettings.instance()    

        # Soil fraction split into: "Rainfed" (previously "Other"), "Forest", "Irrigated".           
        soil = OrderedDict([(name, loadmap(epic_settings.landuse_inputmap[epic_settings.vegetation_landuse[name]])) for name in self.var.prescribed_vegetation])
        if option.get('cropsEPIC'):
            self.var.SoilFraction = xr.DataArray(pd.DataFrame(soil).T, coords=self.var.coord_prescribed_vegetation, dims=self.var.coord_prescribed_vegetation.keys())
        else:
            self.var.SoilFraction = NumpyModified(pd.DataFrame(soil).T.values, dims=self.var.coord_prescribed_vegetation.keys())
        self.var.SoilFraction[0] =  self.var.OtherFraction
        self.var.SoilFraction[1] =  self.var.ForestFraction 
        self.var.SoilFraction[2] =  self.var.IrrigationFraction     
        # Interactive crop fractions (if EPIC is active)
        if option.get('cropsEPIC'):
            self.var.crop_module.setSoilFractions()       
        
    def dynamic(self):
        """ dynamic part of the landusechange module"""
        settings = LisSettings.instance()
        option = settings.options
        binding = settings.binding
        epic_settings = EPICSettings.instance()

        if option['TransientLandUseChange'] and option['readNetcdfStack']:
        
            if option["cropsEPIC"]:
                raise Exception("Land use change for EPIC crops not implemented yet!")
                                
            self.var.ForestFraction = readnetcdf(binding['ForestFractionMaps'], self.var.currentTimeStep(), timestampflag='closest')          #self.var.currentTimeStep()  
            self.var.DirectRunoffFraction = readnetcdf(binding['DirectRunoffFractionMaps'], self.var.currentTimeStep(), timestampflag='closest')          #self.var.currentTimeStep() 
            self.var.WaterFraction = readnetcdf(binding['WaterFractionMaps'], self.var.currentTimeStep(), timestampflag='closest')          #self.var.currentTimeStep() 
            self.var.IrrigationFraction = readnetcdf(binding['IrrigationFractionMaps'], self.var.currentTimeStep(), timestampflag='closest')          #self.var.currentTimeStep() 
            self.var.RiceFraction = readnetcdf(binding['RiceFractionMaps'], self.var.currentTimeStep(), timestampflag='closest')          #self.var.currentTimeStep() 
            self.var.OtherFraction = readnetcdf(binding['OtherFractionMaps'], self.var.currentTimeStep(), timestampflag='closest')          #self.var.currentTimeStep()  
            
            if option['repMBTs']:
                 self.var.ForestFraction_nextstep = []
                 self.var.DirectRunoffFraction_nextstep = []
                 self.var.WaterFraction_nextstep = []
                 self.var.IrrigationFraction_nextstep = []
                 self.var.RiceFraction_nextstep = []
                 self.var.OtherFraction_nextstep = []
                 self.var.DynamicLandCoverDelta = 0.0      
                 self.var.ForestFraction_nextstep = readnetcdf(binding['ForestFractionMaps'], self.var.currentTimeStep()+1, timestampflag='closest')            
                 self.var.DirectRunoffFraction_nextstep = readnetcdf(binding['DirectRunoffFractionMaps'], self.var.currentTimeStep()+1, timestampflag='closest')
                 self.var.WaterFraction_nextstep = readnetcdf(binding['WaterFractionMaps'], self.var.currentTimeStep()+1, timestampflag='closest')
                 self.var.IrrigationFraction_nextstep = readnetcdf(binding['IrrigationFractionMaps'], self.var.currentTimeStep()+1, timestampflag='closest')
                 self.var.RiceFraction_nextstep = readnetcdf(binding['RiceFractionMaps'], self.var.currentTimeStep()+1, timestampflag='closest')
                 self.var.OtherFraction_nextstep = readnetcdf(binding['OtherFractionMaps'], self.var.currentTimeStep()+1, timestampflag='closest')                     
                 self.var.DynamicLandCoverDelta = np.sum(np.abs(self.var.ForestFraction_nextstep-self.var.ForestFraction)+np.abs(self.var.DirectRunoffFraction_nextstep-self.var.DirectRunoffFraction)+np.abs(self.var.WaterFraction_nextstep-self.var.WaterFraction)+np.abs(self.var.IrrigationFraction_nextstep-self.var.IrrigationFraction)+np.abs(self.var.RiceFraction_nextstep-self.var.RiceFraction)+np.abs(self.var.OtherFraction_nextstep-self.var.OtherFraction))
 
            soil = OrderedDict([(name, loadmap(epic_settings.landuse_inputmap[epic_settings.vegetation_landuse[name]])) for name in self.var.prescribed_vegetation])
            if option.get('cropsEPIC'):
                self.var.SoilFraction = xr.DataArray(pd.DataFrame(soil).T, coords=self.var.coord_prescribed_vegetation, dims=self.var.coord_prescribed_vegetation.keys())
            else:
                self.var.SoilFraction = NumpyModified(pd.DataFrame(soil).T.values, dims=self.var.coord_prescribed_vegetation.keys())
            self.var.SoilFraction[0] =  self.var.OtherFraction
            self.var.SoilFraction[1] =  self.var.ForestFraction 
            self.var.SoilFraction[2] =  self.var.IrrigationFraction      
                          
            if not option["cropsEPIC"]: # If EPIC is active, the rice fraction initialisation is handled by EPIC (setSoilFractions in EPIC_main.py)
               self.var.SoilFraction.values[self.var.vegetation.index('Rainfed_prescribed')] += self.var.RiceFraction
                             
            # with EPIC off, rice is treated as other fraction
            # if the fraction of water varies then the other fraction are stored
            self.var.WaterFractionBase = self.var.WaterFraction.copy()
            self.var.OtherFractionBase = self.var.OtherFraction.copy()
            self.var.IrrigationFractionBase = self.var.IrrigationFraction.copy()
            self.var.ForestFractionBase = self.var.ForestFraction.copy()
            self.var.DirectRunoffFractionBase = self.var.DirectRunoffFraction.copy()
            self.var.PermeableFraction = 1 - self.var.DirectRunoffFraction - self.var.WaterFraction   
       



"""

Copyright 2019 European Union

Licensed under the EUPL, Version 1.2 or as soon they will be approved by the European Commission  subsequent versions of the EUPL (the "Licence");

You may not use this work except in compliance with the Licence.
You may obtain a copy of the Licence at:

https://joinup.ec.europa.eu/sites/default/files/inline-files/EUPL%20v1_2%20EN(1).txt

Unless required by applicable law or agreed to in writing, software distributed under the Licence is distributed on an "AS IS" basis,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the Licence for the specific language governing permissions and limitations under the Licence.

"""

from __future__ import print_function, absolute_import, unicode_literals

import os
import inspect
import warnings

from .errors import LisfloodError, LisfloodWarning
from ..hydrological_modules import HydroModule
from ..hydrological_modules import (surface_routing, evapowater, snow, routing, leafarea, inflow, waterlevel,
                                    waterbalance, wateruse, waterabstraction, lakes, riceirrigation, indicatorcalc,
                                    landusechange, frost, groundwater, miscInitial, soilloop, soil,
                                    reservoir, transmission)


class ModulesInputs:
    root_package = 'lisflood.hydrological_modules'
    # dict representing modules activated per option
    lisflood_modules = {
        # not optional modules
        'all': [surface_routing, snow, routing, leafarea, landusechange,
                frost, groundwater, miscInitial, soil],
        # list of modules activated by options in LisFlood settings.xml
        'inflow': [inflow],
        'wateruse': [wateruse],
        'groundwaterSmooth': [waterabstraction],
        'wateruseRegion': [waterabstraction],
        'drainedIrrigation': [soilloop, soil],
        'riceIrrigation': [riceirrigation, waterabstraction],
        'indicator': [lakes, indicatorcalc, waterabstraction],
        'openwaterevapo': [evapowater],
        'varfractionwater': [evapowater],
        'TransientLandUseChange': [landusechange, indicatorcalc, waterabstraction],
        'simulateLakes': [lakes, indicatorcalc, routing, waterabstraction, waterbalance],
        'simulateReservoirs': [reservoir, indicatorcalc, routing, waterabstraction, waterbalance],
        'simulatePF': [soilloop, soil],
        'simulateWaterLevels': [waterlevel],
        'TransLoss': [transmission],
        'gridSizeUserDefined': [miscInitial],
    }

    @classmethod
    def check(cls, settings):
        """
        :param settings: LisSettings object
        """
        binding = settings.binding
        successful_checks = 0
        all_errors = []
        total_checks = 1  # at least check PathOut
        # First check PathOut
        if not (os.path.exists(settings.output_dir) and os.path.isdir(settings.output_dir) and os.access(settings.output_dir, os.W_OK)):
            msg = '\n\nPath defined in PathOut is not writable: {}'.format(settings.output_dir)
            warnings.warn(LisfloodWarning(msg))
            all_errors.append(msg)
        else:
            successful_checks += 1

        for option, modules in cls.lisflood_modules.items():
            if not (settings.options.get(option) or option == 'all'):
                # module is not activated
                continue
            for hydro_module in modules:
                clzzs = [obj for n, obj in inspect.getmembers(hydro_module)
                         if inspect.isclass(obj) and issubclass(obj, HydroModule)
                         and obj is not HydroModule and str(obj.__module__) == hydro_module.__name__]
                total_checks += len(clzzs)
                for clz in clzzs:
                    errors = clz.check_input_files(option)
                    successful_checks += int(bool(not errors))
                    all_errors += errors
        if successful_checks < total_checks:
            # some checks failed
            raise LisfloodError(
                '\n\nMissing files or misconfigured paths to run LisFlood, according activated modules. '
                'Please check your settings file {}.\n{}'.format(settings.settings_path, all_errors)
            )


class MeteoForcings:
    @classmethod
    def check(cls, settings):
        # TODO implement
        binding = settings.binding
        pass



"""

Copyright 2019 European Union

Licensed under the EUPL, Version 1.2 or as soon they will be approved by the European Commission  subsequent versions of the EUPL (the "Licence");

You may not use this work except in compliance with the Licence.
You may obtain a copy of the Licence at:

https://joinup.ec.europa.eu/sites/default/files/inline-files/EUPL%20v1_2%20EN(1).txt

Unless required by applicable law or agreed to in writing, software distributed under the Licence is distributed on an "AS IS" basis,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the Licence for the specific language governing permissions and limitations under the Licence.

"""

from __future__ import absolute_import, print_function

import datetime

from pcraster import areatotal, cover, ifthen, upstream
import numpy as np

from ..global_modules.add1 import loadmap, readnetcdf, decompress, compressArray
from ..global_modules.settings import LisSettings, MaskInfo
from . import HydroModule


class indicatorcalc(HydroModule):

    """
    # ************************************************************
    # ***** Indicator calculation ************************************
    # ************************************************************
    """
    input_files_keys = {'indicator': ['Population', 'LandUseMask'],
                        'TransientLandUseChange': ['PopulationMaps']}
    module_name = 'IndicatorCalculation'

    def __init__(self, indicatorcalc_variable):
        self.var = indicatorcalc_variable

# --------------------------------------------------------------------------
# --------------------------------------------------------------------------

    def initial(self):
        """ initial part of the indicator calculation module
        """

        self.var.monthend = False
        self.var.yearend = False
        settings = LisSettings.instance()
        binding = settings.binding
        option = settings.options
        maskinfo = MaskInfo.instance()
        try:
            self.var.DefineEndofYear = int(binding['DefineEndofYear'])
        except (ValueError, KeyError):
            self.var.DefineEndofYear = 365

        if option['indicator'] and option['wateruse']:
            self.var.Population = loadmap('Population') # population per pixel
            self.var.LandUseMask = loadmap('LandUseMask')
            self.var.RegionPopulation = np.take(np.bincount(self.var.WUseRegionC, weights=self.var.Population), self.var.WUseRegionC)
            self.dynamic_setzero()
            
            self.var.DayCounter = 0           
            self.var.MonthETpotMM = maskinfo.in_zero() 
            self.var.MonthETactMM = maskinfo.in_zero() 
           
            

    def dynamic(self):
        """ dynamic part of the indicator calculation module
        """
        settings = LisSettings.instance()
        binding = settings.binding
        option = settings.options
        maskinfo = MaskInfo.instance()
        if option['TransientLandUseChange']:
            self.var.Population = readnetcdf(binding['PopulationMaps'], self.var.currentTimeStep())
            self.var.RegionPopulation = np.take(np.bincount(self.var.WUseRegionC, weights=self.var.Population), self.var.WUseRegionC)
            # population sum in Regions

        if option['wateruse'] and option['indicator']:
            # check if it is the last monthly or annual time step
            next_date_time = self.var.CalendarDate + datetime.timedelta(seconds=int(binding["DtSec"]))
            self.var.monthend = next_date_time.month != self.var.CalendarDate.month
            self.var.yearend = next_date_time.year != self.var.CalendarDate.year
            # sum up every day
            self.var.DayCounter += 1
            
            self.var.MonthETpotMM = self.var.MonthETpotMM + self.var.ETRef 
            self.var.MonthETactMM = self.var.MonthETactMM + self.var.deffraction(self.var.TaInterception) + self.var.TaPixel + self.var.ESActPixel 

            if option['openwaterevapo']: 
                self.var.MonthETactMM += self.var.EvaAddM3 * self.var.M3toMM 

            self.var.MonthETdifMM = np.maximum((self.var.MonthETpotMM - self.var.MonthETactMM)*self.var.LandUseMask, maskinfo.in_zero())  
            # ; land use mask can be used to mask out deserts and high mountains, where no agriculture is possible            

            self.var.MonthAbstractionRequiredAllSourcesM3 += self.var.abstraction_allSources_required_M3
            self.var.MonthAbstractionRequiredSurfaceGroundWaterM3 += self.var.abstraction_SwGw_required_M3
            self.var.MonthAbstractionRequiredSurfaceWaterM3 += self.var.abstraction_SW_required_M3
            self.var.MonthConsumptionRequiredM3 += self.var.consumption_SwGw_required_M3
            self.var.MonthConsumptionActualM3 += self.var.consumption_SwGw_actual_M3
            self.var.MonthDisM3     =	self.var.MonthDisM3 + self.var.ChanQAvg * self.var.DtSec

            self.var.MonthWaterAbstractedfromLakesReservoirsM3 = self.var.MonthWaterAbstractedfromLakesReservoirsM3 + self.var.ReservoirAbstractionM3 + self.var.LakeAbstractionM3
            self.var.RegionMonthIrrigationShortageM3 = self.var.RegionMonthIrrigationShortageM3 + self.var.areatotal_shortage_SW_M3
            self.var.MonthInternalFlowM3 = self.var.MonthInternalFlowM3 + self.var.ToChanM3Runoff
            
            if self.var.monthend:

                # INTERNAL FLOW
                # available LakeStorageM3 and ReservoirStorageM3 for potential abstraction at end of month in region  
                if option['simulateLakes'] and option['simulateReservoirs']:            
                   self.var.RegionMonthReservoirAndLakeStorageM3 = np.take(np.bincount(self.var.WUseRegionC,weights=self.var.ReservoirStorageM3+self.var.LakeStorageM3),self.var.WUseRegionC)
                   # monthly abstraction from lakes and reservoirs
                   self.var.RegionMonthWaterAbstractedfromLakesReservoirsM3 = np.take(np.bincount(self.var.WUseRegionC,weights=self.var.MonthWaterAbstractedfromLakesReservoirsM3),self.var.WUseRegionC)
                
                self.var.RegionMonthInternalFlowM3 = np.take(np.bincount(self.var.WUseRegionC,weights=self.var.MonthInternalFlowM3),self.var.WUseRegionC)
                   # note Reservoir and Lake storage need to be taken into account seperately

                # EXTERNAL FLOW
                wreg = decompress(self.var.WUseRegionC)
                     # to pcraster map because of the following expression!
                self.var.RegionMonthExternalInflowM3 = compressArray(areatotal(cover(ifthen(self.var.WaterRegionInflowPoints != 0, upstream(self.var.LddStructuresKinematic,decompress(self.var.MonthDisM3))),0),wreg))
                self.var.RegionMonthAbstractionRequiredAllSourcesM3 =  np.take(np.bincount(self.var.WUseRegionC,weights=self.var.MonthAbstractionRequiredAllSourcesM3),self.var.WUseRegionC)
                self.var.RegionMonthAbstractionRequiredSurfaceGroundWaterM3 = np.take(np.bincount(self.var.WUseRegionC,weights=self.var.MonthAbstractionRequiredSurfaceGroundWaterM3),self.var.WUseRegionC)
                self.var.RegionMonthAbstractionRequiredSurfaceWaterM3 = np.take(np.bincount(self.var.WUseRegionC,weights=self.var.MonthAbstractionRequiredSurfaceWaterM3),self.var.WUseRegionC)
                self.var.RegionMonthConsumptionRequiredM3 = np.take(np.bincount(self.var.WUseRegionC,weights=self.var.MonthConsumptionRequiredM3),self.var.WUseRegionC)
                self.var.RegionMonthConsumptionActualM3 = np.take(np.bincount(self.var.WUseRegionC,weights=self.var.MonthConsumptionActualM3),self.var.WUseRegionC)


                # Calculation of WEI: everything in m3, totalled over the water region

                self.var.UpstreamInflowM3 = self.var.RegionMonthExternalInflowM3
                self.var.LocalFreshwaterM3 = self.var.RegionMonthInternalFlowM3
                # still to be decided if reservoir and lake water availability is added here
                self.var.LocalTotalWaterDemandM3 = self.var.RegionMonthAbstractionRequiredAllSourcesM3
                RemainingDemandM3 = np.maximum(self.var.LocalTotalWaterDemandM3 - self.var.LocalFreshwaterM3,0.0)
                # this is the demand that cannot be met by local water supply
                self.var.UpstreamInflowUsedM3 = np.minimum(RemainingDemandM3,self.var.UpstreamInflowM3)
                # the amount of upstream water an area really depends on
                self.var.FossilGroundwaterUsedM3 = np.maximum(RemainingDemandM3 - self.var.UpstreamInflowUsedM3,0.0)
                # the amount of water that cannot be met locally nor with international water
                # likely this is the amount of water drawn from deep groundwater
                freshwater_total = self.var.UpstreamInflowM3 + self.var.LocalFreshwaterM3
                is_freshwater_available = freshwater_total > 0
                self.var.WEI_Dem = np.where(is_freshwater_available, self.var.RegionMonthAbstractionRequiredAllSourcesM3 / freshwater_total, 0.)
                self.var.WEI_Abs = np.where(is_freshwater_available, self.var.RegionMonthAbstractionRequiredSurfaceGroundWaterM3 / freshwater_total, 0.)
                self.var.WEI_Cns = np.where(is_freshwater_available, self.var.RegionMonthConsumptionRequiredM3 / freshwater_total, 0.)
                self.var.WEI_Plus = np.where(is_freshwater_available, self.var.RegionMonthConsumptionActualM3 / freshwater_total, 0.)
                ## TEST
                #print('WEI_D: {}, WEI_+: {}'.format(self.var.WEI_Dem.mean(), self.var.WEI_Plus.mean()))

                self.var.WaterSustainabilityIndex =  np.where(self.var.LocalTotalWaterDemandM3 > 0,
                                                              self.var.FossilGroundwaterUsedM3 / (self.var.LocalTotalWaterDemandM3+1),0.0)
                # De Roo 2015: WTI, if above zero, indicates that situation is unsustainable
                # if index is 0 means sustainable situtation: no groundwater or desalination water used
                # if index is 1 means area relies completely on groundwater or desalination water
                # the '+1' is to prevent division by small values, leading to very large and misleading indicator values

                self.var.WaterDependencyIndex =  np.where(self.var.LocalTotalWaterDemandM3 > 0, self.var.UpstreamInflowUsedM3 / (self.var.LocalTotalWaterDemandM3+1),0)
                # De Roo 2015: WDI, dependency on upstreamwater, as a fraction of the total local demand
                # the '+1' is to prevent division by small values, leading to very large and misleading indicator values

                self.var.WaterSecurityIndex = np.where(self.var.UpstreamInflowM3 > 0,  self.var.UpstreamInflowUsedM3 / (self.var.UpstreamInflowM3+1),0)
                # De Roo 2015: WSI, indicates the vulnerability to the available upstream water;
                # if only 10% of upstream inflow is use, WSI would be 0.1 indicating low vulnerability
                # if WSI is close to 1, situation is very vulnerable
                # the '+1' is to prevent division by small values, leading to very large and misleading indicator values

                self.var.FalkenmarkM3Capita1 =  np.where(self.var.RegionPopulation > 0.0,self.var.RegionMonthInternalFlowM3*12/self.var.RegionPopulation,0)
                self.var.FalkenmarkM3Capita2 =  np.where(self.var.RegionPopulation > 0.0,self.var.LocalFreshwaterM3*12/self.var.RegionPopulation,0)
                self.var.FalkenmarkM3Capita3 =  np.where(self.var.RegionPopulation > 0.0,freshwater_total*12/self.var.RegionPopulation,0)
                if option['repWaterUse']:
                    # Required consumptions (from surface water and groundwater) by sector, including actual figure for irrigation (month, region)
                    self.var.consumption_required_domestic_M3MonthRegion = np.take(np.bincount(self.var.WUseRegionC,weights=self.var.consumption_required_domestic_MM_month),self.var.WUseRegionC)
                    self.var.consumption_required_energy_M3MonthRegion =  np.take(np.bincount(self.var.WUseRegionC,weights=self.var.consumption_required_energy_MM_month * self.var.MMtoM3),self.var.WUseRegionC)
                    self.var.consumption_required_industry_M3MonthRegion = np.take(np.bincount(self.var.WUseRegionC,weights=self.var.consumption_required_industry_MM_month * self.var.MMtoM3),self.var.WUseRegionC)
                    self.var.consumption_required_livestock_M3MonthRegion = np.take(np.bincount(self.var.WUseRegionC,weights=self.var.consumption_required_livestock_MM_month *self.var.MMtoM3),self.var.WUseRegionC)
                    self.var.consumption_required_irrigation_M3MonthRegion = np.take(np.bincount(self.var.WUseRegionC,weights=self.var.consumption_required_irrigation_MM_month * self.var.MMtoM3),self.var.WUseRegionC)
                    self.var.consumption_actual_irrigation_M3MonthRegion = np.take(np.bincount(self.var.WUseRegionC,weights=self.var.consumption_actual_irrigation_MM_month *self.var.MMtoM3),self.var.WUseRegionC)
                    # Irrigation abstraction (month, self.var.WUseRegionC), region)
                    self.var.abstraction_allSources_required_irrigation_M3MonthRegion =np.take(np.bincount(self.var.WUseRegionC,weights=self.var.abstraction_allSources_required_irrigation_M3Month),self.var.WUseRegionC)
                    self.var.abstraction_allSources_actual_irrigation_M3MonthRegion =np.take(np.bincount(self.var.WUseRegionC,weights=self.var.abstraction_allSources_actual_irrigation_M3Month),self.var.WUseRegionC)
                    self.var.abstraction_SwGw_required_irrigation_M3MonthRegion =np.take(np.bincount(self.var.WUseRegionC,weights=self.var.abstraction_SwGw_required_irrigation_M3Month),self.var.WUseRegionC)
                    self.var.abstraction_SwGw_actual_irrigation_M3MonthRegion =np.take(np.bincount(self.var.WUseRegionC,weights=self.var.abstraction_SwGw_actual_irrigation_M3Month),self.var.WUseRegionC)

         
# --------------------------------------------------------------------------
    def dynamic_setzero(self):
        """ dynamic part of the indicator calculation module
            which set the monthly (yearly) values back to start
        """
        settings = LisSettings.instance()
        option = settings.options
        maskinfo = MaskInfo.instance()  

        self.var.DayCounter = 0
        self.var.MonthETpotMM = maskinfo.in_zero() 
        self.var.MonthETactMM = maskinfo.in_zero()  
        self.var.MonthAbstractionRequiredAllSourcesM3 =  maskinfo.in_zero()
        self.var.MonthAbstractionRequiredSurfaceGroundWaterM3 =  maskinfo.in_zero()
        self.var.MonthAbstractionRequiredSurfaceWaterM3 =  maskinfo.in_zero()
        self.var.MonthConsumptionActualM3 =  maskinfo.in_zero()
        self.var.MonthConsumptionRequiredM3 =  maskinfo.in_zero()
        self.var.MonthDisM3 =  maskinfo.in_zero()
        self.var.MonthInternalFlowM3 =  maskinfo.in_zero()
        self.var.MonthExternalInflowM3 = maskinfo.in_zero()
        self.var.RegionMonthIrrigationShortageM3 = maskinfo.in_zero()
        self.var.MonthWaterAbstractedfromLakesReservoirsM3 =  maskinfo.in_zero()

        # Required consumptions (from surface water and groundwater) by sector, including actual figure for irrigation (month, region)
        ##self.var.consumption_required_domestic_MM_month = np.zeros(self.var.num_pixel)
        self.var.consumption_required_domestic_MM_month =  np.zeros(self.var.num_pixel) 
        self.var.consumption_required_energy_MM_month = np.zeros(self.var.num_pixel)
        self.var.consumption_required_industry_MM_month = np.zeros(self.var.num_pixel)
        self.var.consumption_required_livestock_MM_month = np.zeros(self.var.num_pixel)
        self.var.consumption_required_irrigation_MM_month = np.zeros(self.var.num_pixel)
        self.var.consumption_actual_irrigation_MM_month = np.zeros(self.var.num_pixel)
        # Irrigation abstraction (month, region)
        self.var.abstraction_allSources_required_irrigation_M3Month = np.zeros(self.var.num_pixel)
        self.var.abstraction_allSources_actual_irrigation_M3Month = np.zeros(self.var.num_pixel)
        self.var.abstraction_SwGw_required_irrigation_M3Month = np.zeros(self.var.num_pixel)
        self.var.abstraction_SwGw_actual_irrigation_M3Month = np.zeros(self.var.num_pixel)                                 

        
            



"""
Copyright 2019 European Union

Licensed under the EUPL, Version 1.2 or as soon they will be approved by the European Commission  subsequent versions of the EUPL (the "Licence");

You may not use this work except in compliance with the Licence.
You may obtain a copy of the Licence at:

https://joinup.ec.europa.eu/sites/default/files/inline-files/EUPL%20v1_2%20EN(1).txt

Unless required by applicable law or agreed to in writing, software distributed under the Licence is distributed on an "AS IS" basis,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the Licence for the specific language governing permissions and limitations under the Licence.

---------------------------------------------------------------------------------------------------------------------------------------
Use python setup.y publish to publish versioned tags and pypi package

IMPORTANT Note:
To test pip installation:
python setup.py testpypi
pip install --index-url https://test.pypi.org/simple/ lisflood-model==2.8.14

To publish on PyPi:

python setup.py publish

It's equivalent to:

1. python setup.py sdist
2. twine upload dist/*
3. git tag {version}
4. git push --tags

Install

pip install lisflood-model
"""

import os
import subprocess
import sys
from glob import glob
from shutil import rmtree

from setuptools import find_packages, Extension, setup, Command, __path__ as setuppath

pip_package = os.path.normpath(os.path.join(setuppath[0], '../lisflood'))
if os.path.exists(pip_package):
    print('[-] Removing current package installed {}'.format(pip_package))
    rmtree(pip_package)

current_dir = os.path.dirname(os.path.abspath(__file__))
sys.path.append(os.path.join(current_dir, './src/'))

from lisflood import __authors__
version_file = os.path.join(current_dir, 'VERSION')

with open(version_file, 'r') as f:
    version = f.read().strip()

print(">>>>>>>>>>>>>>>> Building LISFLOOD version {} <<<<<<<<<<<<<<<<<<".format(version))
try:
    # noqa
    from Cython.Build import cythonize
    HAS_CYTHON = True
except ImportError:
    HAS_CYTHON = False

src_ext = 'src/lisflood/hydrological_modules/kinematic_wave_parallel_tools.{}'
extension_ext = 'pyx' if HAS_CYTHON else 'c'

ext_modules = [
    Extension(
        'lisflood.hydrological_modules.kinematic_wave_parallel_tools',
        [src_ext.format(extension_ext)],
        extra_compile_args=['-O3', "-march=native", '-fopenmp'],
        extra_link_args=['-fopenmp']
    )
]

if HAS_CYTHON:
    ext_modules = cythonize(ext_modules)

readme_file = os.path.join(current_dir, 'README.md')
with open(readme_file, 'r') as f:
    long_description = f.read()


class UploadCommand(Command):
    """Support setup.py upload."""

    description = 'Publish lisflood-model package.'
    user_options = []

    @staticmethod
    def print_console(s):
        print('\033[1m{0}\033[0m'.format(s))

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        try:
            self.print_console('Removing previous builds...')
            rmtree(os.path.join(current_dir, 'dist'))
        except OSError:
            pass

        self.print_console('Building Source and Wheel (universal) distribution...')
        os.system('{} setup.py sdist'.format(sys.executable))

        self.print_console('Uploading the package to PyPI via Twine...')
        os.system('twine upload dist/*')

        self.print_console('Pushing git tags...')
        os.system('git tag v{}'.format(version))
        os.system('git push --tags')

        sys.exit()


class UploadCommandTest(UploadCommand):

    def run(self):
        try:
            self.print_console('Removing previous builds...')
            rmtree(os.path.join(current_dir, 'dist'))
        except OSError:
            pass

        self.print_console('Building Source and Wheel (universal) distribution...')
        os.system('{} setup.py sdist'.format(sys.executable))

        self.print_console('Uploading the package to test PyPI via Twine...')
        os.system('twine upload --repository testpypi dist/*')

        sys.exit()


def _get_gdal_version():
    try:
        p = subprocess.Popen(['gdal-config', '--version'], stdout=subprocess.PIPE)
    except FileNotFoundError:
        raise SystemError('gdal-config not found.'
                          'GDAL seems not installed. '
                          'Please, install GDAL binaries and libraries for your system '
                          'and then install the relative pip package.')
    else:
        return p.communicate()[0].splitlines()[0].decode()


gdal_version = _get_gdal_version()
req_file = 'requirements.txt'
requirements = [l for l in open(req_file).readlines() if l and not l.startswith('#')]
requirements += ['GDAL=={}'.format(gdal_version)]
setup(
    name='lisflood-model',
    version=version,
    package_dir={'': 'src/'},
    py_modules=[os.path.splitext(os.path.basename(path))[0] for path in glob('src/*.py*')],
    include_package_data=True,
    data_files=[('', ['VERSION']), ('settings', ['src/lisfloodSettings_reference.xml'])],
    packages=find_packages('src'),
    long_description=long_description,
    long_description_content_type='text/markdown',
    description='LISFLOOD model python module',
    author=__authors__,
    author_email='carlo.russo@ext.ec.europa.eu',
    keywords=['lisflood', 'lisvap', 'efas', 'glofas', 'copernicus', 'ecmwf'],
    license='EUPL 1.2',
    url='https://github.com/ec-jrc/lisflood-code',
    download_url='https://github.com/ec-jrc/lisflood-code/archive/{}.tar.gz'.format(version),
    setup_requires=[
            'nine',
            'setuptools>=41.0',
            'numpy>=1.15',
            'cython',
    ],
    install_requires=requirements,
    scripts=['bin/lisflood'],
    ext_modules=ext_modules,
    zip_safe=True,
    classifiers=[
        # complete classifier list: http://pypi.python.org/pypi?%3Aaction=list_classifiers
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Intended Audience :: Education',
        'Intended Audience :: Financial and Insurance Industry',
        'Intended Audience :: Other Audience',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: European Union Public Licence 1.2 (EUPL 1.2)',
        'Operating System :: Unix',
        'Operating System :: POSIX',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: MacOS :: MacOS X',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.7',
        'Topic :: Scientific/Engineering :: Physics',
    ],
    # setup.py publish to pypi.
    cmdclass={
        'upload': UploadCommand,
        'publish': UploadCommand,
        'testpypi': UploadCommandTest,
    },
)



"""

Copyright 2019 European Union

Licensed under the EUPL, Version 1.2 or as soon they will be approved by the European Commission  subsequent versions of the EUPL (the "Licence");

You may not use this work except in compliance with the Licence.
You may obtain a copy of the Licence at:

https://joinup.ec.europa.eu/sites/default/files/inline-files/EUPL%20v1_2%20EN(1).txt

Unless required by applicable law or agreed to in writing, software distributed under the Licence is distributed on an "AS IS" basis,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the Licence for the specific language governing permissions and limitations under the Licence.

"""

from functools import wraps
import copy


def counted(fn):
    def wrapper(*args, **kwargs):
        wrapper.called += 1
        return fn(*args, **kwargs)

    wrapper.called = 0
    wrapper.__name__ = fn.__name__
    return wrapper


def cached(f):
    """
    Simple cache for small objects like parsing options
    """

    _cache = {}

    @wraps(f)
    def _decorator(args):
        args = tuple(args)
        if args not in _cache:
            _cache[args] = f(args)
        return _cache[args]

    return _decorator


class Cache:
    """
    Class decorator used to cache large objects read from disk
    Mostly used for forcings and static maps
    """

    cache = {}
    found = {}

    def __init__(self, fn):
        self.name = fn.__name__
        self.fn = fn
        # we need to put the counter in a dict
        # or we lose the reference
        if self.name not in self.found:
            self.found[self.name] = 0

    def __call__(self, *args, **kwargs):

        key = '{}, {}, {}'.format(self.name, args, kwargs)

        if key not in self.cache:
            data = self.fn(*args, **kwargs)
            # we don't cache small objects (e.g. floats from loadmap)
            if isinstance(data, float):
                return_data = data
            else:
                self.cache[key] = data
                return_data = self.cache[key]
        else:
            return_data = self.cache[key]
            self.found[self.name] += 1
        return return_data
    
    @classmethod
    def clear(cls):
        print('Clearing cache')
        cls.cache.clear()
        for i in cls.found:
            cls.found[i] = 0

    @classmethod
    def size(cls):
        return len(cls.cache)

    @classmethod
    def extract(cls):
        return copy.deepcopy(cls.cache)

    @classmethod
    def apply(cls, cache_in):
        # We need to loop to keep the reference to cache
        for i in cache_in:
            cls.cache[i] = cache_in[i]

    @classmethod
    def values_found(cls):
        return sum(cls.found.values())

    @classmethod
    def info(cls):
        print('Caching')
        print('Number of items cached: {}'.format(cls.size()))
        print('Number of items retrieved: {}'.format(cls.found))
        print('Keys:')
        for key in cls.cache.keys():
            print('   - {}'.format(key))



"""

Copyright 2019 European Union

Licensed under the EUPL, Version 1.2 or as soon they will be approved by the European Commission  subsequent versions of the EUPL (the "Licence");

You may not use this work except in compliance with the Licence.
You may obtain a copy of the Licence at:

https://joinup.ec.europa.eu/sites/default/files/inline-files/EUPL%20v1_2%20EN(1).txt

Unless required by applicable law or agreed to in writing, software distributed under the Licence is distributed on an "AS IS" basis,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the Licence for the specific language governing permissions and limitations under the Licence.

"""
from __future__ import absolute_import, print_function

import numpy as np

from lisflood.global_modules.errors import LisfloodError

from ..global_modules.add1 import loadmap
from ..global_modules.settings import MaskInfo, LisSettings
from . import HydroModule


class riceirrigation(HydroModule):
    """
    # ************************************************************
    # ***** Rice irrigation   ************************************
    # ************************************************************
    """
    input_files_keys = {'riceIrrigation': ['RiceFlooding', 'RicePercolation', 'RicePlantingDay1',
                                           'RiceHarvestDay1', 'RicePlantingDay2', 'RiceHarvestDay2']}
    module_name = 'RiceIrrigation'

    def __init__(self, riceirrigation_variable):
        self.var = riceirrigation_variable

    # --------------------------------------------------------------------------
    # --------------------------------------------------------------------------

    def initial(self):
        """ initial part of the rice irrigation module
        """
        maskinfo = MaskInfo.instance()
        self.var.PaddyRiceWaterAbstractionFromSurfaceWaterM3 = maskinfo.in_zero()
        settings = LisSettings.instance()
        option = settings.options
        if option['riceIrrigation']:
            if not option['wateruse']:
                # flag with an error message and STOP:
                # riceIrrigation ON MUST HAVE wateruse option ON
                msg = "riceIrrigation module ON MUST HAVE wateruse option ON in setting file"
                raise LisfloodError(msg)

            # ************************************************************
            # ***** PADDY RICE IRRIGATION AND ABSTRACTION ******************
            # ************************************************************

            # Additional water for paddy rice cultivation is calculated seperately, as well as additional open water evaporation from rice fields
            self.var.RiceFlooding = loadmap('RiceFlooding')
            # 10 mm for 10 days (total 10cm water)
            self.var.RicePercolation = loadmap('RicePercolation')
            # FAO: percolation for heavy clay soils: PERC = 2 mm/day

            self.var.RicePlantingDay1 = loadmap('RicePlantingDay1')
            # starting day (of the year) of first rice planting
            self.var.RiceHarvestDay1 = loadmap('RiceHarvestDay1')
            # starting day (of the year) of first rice harvest

            self.var.RicePlantingDay2 = loadmap('RicePlantingDay2')
            # starting day (of the year) of second rice planting
            self.var.RiceHarvestDay2 = loadmap('RiceHarvestDay2')
            # starting day (of the year) of 2nd rice harvest

    def dynamic(self):
        """ dynamic part of the rice irrigation routine
           inside the water abstraction routine
        """
        settings = LisSettings.instance()
        option = settings.options
        maskinfo = MaskInfo.instance()
        if option['riceIrrigation']:
            veg = "Rainfed_prescribed" # ONCE RICE IS SIMULATED IN EPIC, THIS MODULE SHOULD BE SKIPPED WHEN EPIC IS ON!
            iveg, ilanduse, _ = self.var.get_landuse_and_indexes_from_vegetation_GLOBAL(veg)
 
            # water needed for paddy rice is assumed to consist of:
            # phase 1: field preparation: soil saturation (assumed to happen in 10 days, 20 days before planting)
            # phase 2: flood fields (assumed to happen in 10 days, 10 days before planting)
            # phase 3: planting, while keep constant water level during growing season (open water evaporation)
            # phase 4: stop keeping constant water level 20 days before harvest date
            # phase 5: start draining 10 days before harvest date
            
            # ORIGINAL
            #RiceSoilSaturationDemandM3 = (self.var.WS1[ilanduse] - self.var.W1[iveg] + self.var.WS2[ilanduse] - self.var.W2[iveg]) * self.var.RiceFraction * self.var.MMtoM3 * self.var.DtDay
            # this part is using the whole other fraction to calculate the demand -> an rice only soil part is needed
            # RiceIrrigationDemandM3 unit is m3 per time interval [m3/dt]
            
            # EDITED on Jan 15th 2022: saturation demand is computed considering only soil layers 1a and 1b (NO 2)
            RiceSoilSaturationDemandM3 = (self.var.WS1.values[ilanduse] - self.var.W1.values[iveg]) * self.var.RiceFraction * self.var.MMtoM3 * self.var.DtDay
            # this part is using the whole other fraction to calculate the demand -> an rice only soil part is needed
            # RiceIrrigationDemandM3 unit is m3 per time interval [m3/dt]            

            pl_20 = self.var.RicePlantingDay1 - 20
            pl_20 = np.where(pl_20 < 0, 365 + pl_20, pl_20)
            pl_10 = self.var.RicePlantingDay1 - 10
            pl_10 = np.where(pl_10 < 0, 365 + pl_10, pl_10)

            ha_20 = self.var.RiceHarvestDay1 - 20
            ha_20 = np.where(ha_20 < 0, 365 + ha_20, ha_20)
            ha_10 = self.var.RiceHarvestDay1 - 10
            ha_10 = np.where(ha_10 < 0, 365 + ha_10, ha_10)
            
 
            # for Europe ok, but for Global planting can be on the 330 and harvest on the 90, so harvest < planting
            # or riceplanting = 5 => riceplanting -20 =350 ==> riceplanting < riceplanting -20

            """ phase 1: field preparation: soil saturation (assumed to happen in 10 days, 20 days before planting)"""
            RiceSoilSaturationM3 = np.where((self.var.CalendarDay >= pl_20) & (self.var.CalendarDay < pl_10),
                                            0.1 * RiceSoilSaturationDemandM3, maskinfo.in_zero())

            RiceEva = np.maximum(self.var.EWRef - (self.var.ESAct.values[iveg] + self.var.Ta.values[iveg]), 0)            
            RiceEvaporationDemandM3 = RiceEva * self.var.RiceFraction * self.var.MMtoM3  # m3 per time interval
            # should not happen, but just to be sure that this doesnt go <0
            # part of the evaporation is already taken out in soil module!
            # substracting the soil evaporation and transpiration which was already taken off in the soil module
            
            RiceFloodingDemandM3 = self.var.RiceFlooding * self.var.RiceFraction * self.var.MMtoM3 * self.var.DtDay  # m3 per time interval

            """ phase 2: flood fields (assumed to happen in 10 days, 10 days before planting)"""

            RiceFloodingM3 = np.where(
                (self.var.CalendarDay >= pl_10) & (self.var.CalendarDay < self.var.RicePlantingDay1),
                RiceFloodingDemandM3 + RiceEvaporationDemandM3, maskinfo.in_zero())  # m3 per time interval
            # part of the evaporation is already taken out in soil module!
            # assumption is that a fixed water layer is kept on the rice fields, totalling RiceFlooding*10 in mmm (typically 50 or 100 mm)
            # application is spread out over 10 days
            # open water evaporation at the same time

            """ phase 3: planting, while keep constant water level during growing season (open water evaporation) """
            RiceEvaporationM3 = np.where(
                (self.var.CalendarDay >= self.var.RicePlantingDay1) & (self.var.CalendarDay < ha_20),
                RiceEvaporationDemandM3, maskinfo.in_zero())  # m3 per time interval
            # substracting the soil evaporation which was already taken off in the soil module (also transpitation should be tyaken off )

            RicePercolationDemandM3 = self.var.RicePercolation * self.var.RiceFraction * self.var.MMtoM3 * self.var.DtDay  # m3 per time interval         
            RicePercolationM3 = np.where(
                (self.var.CalendarDay >= self.var.RicePlantingDay1) & (self.var.CalendarDay < ha_20),
                RicePercolationDemandM3, maskinfo.in_zero())  # m3 per time interval
            # FAO: percolation for heavy clay soils: PERC = 2 mm/day
            
            self.var.PaddyRiceWaterAbstractionFromSurfaceWaterM3 = RiceSoilSaturationM3 + RiceFloodingM3 + RiceEvaporationM3 + RicePercolationM3  # m3 per time interval
            # m3 water needed for paddyrice

            """# phase 4: stop keeping constant water level 20 days before harvest date
                 phase 5: start draining 10 days before harvest date"""
            # RiceDrainageM3=if((CalendarDay ge (RiceHarvestDay1-10)) and (CalendarDay le RiceHarvestDay1),(WS1-WFC1+WS2-WFC2)*RiceFraction*MMtoM3,0)

            # ORIGINAL: DRAINAGE from all the soil layers
            #RiceDrainageDemandM3 = (self.var.WS1[ilanduse] - self.var.WFC1[ilanduse] + self.var.WS2[ilanduse] - \
            #   self.var.WFC2[ilanduse]) * self.var.RiceFraction * self.var.MMtoM3 * self.var.DtDay  # m3 per time interval
            
            # EDITED on January 15th 2022: DRAINAGE from layers 1a and 1b (NO 2)    
            RiceDrainageDemandM3 = (self.var.WS1.values[ilanduse] - self.var.WFC1.values[ilanduse]) * self.var.RiceFraction * self.var.MMtoM3 * self.var.DtDay  # m3 per time interval    
            
            RiceDrainageM3 = np.where(
                (self.var.CalendarDay >= ha_10) & (self.var.CalendarDay < self.var.RiceHarvestDay1),
                0.1 * RiceDrainageDemandM3, maskinfo.in_zero())

            # drainage until FC to soil/groundwater at end of season
            # assumption that the last weeks before harvest the 50mm water layer is completely evaporating
            # needs to be transported to channel system or being drained

            self.var.UZ.values[iveg][:] += np.where(self.var.SoilFraction.values[iveg] > 0.0,
                                       (RiceDrainageM3 + RicePercolationM3) * self.var.M3toMM / self.var.SoilFraction.values[iveg], ## OtherFraction
                                       0.0)
            # drained water is added to Upper Zone
            